###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Integration with external services for Production Management System"""

from __future__ import annotations

import asyncio
import base64
import binascii
import hashlib
import json
import os
import re
import time
from abc import ABCMeta
from collections import defaultdict
from datetime import datetime, timezone
from pathlib import Path

import gitlab
import prompt_toolkit
import requests
import yaml
from bs4 import BeautifulSoup
from rich.progress import Progress, MofNCompleteColumn, TimeElapsedColumn
from pydantic import BaseModel as _BaseModel, Field
from rich.prompt import Prompt

from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

CACHE_DIR = Path.home() / ".cache" / "dirac" / "production-management"
METADATA_HEADER = "<details><summary>Prod Metadata</summary>"
RE_METADATA = re.compile(r"(<details>\n?<summary>Prod Metadata</summary>\n\n```json\n)([^`]+)(\n```\n\n</details>)")
PROGRESS_COLUMNS = [MofNCompleteColumn(), *Progress.get_default_columns(), TimeElapsedColumn()]

ISSUE_STATE_RESPONSIBLE = {
    "staging": "Computing",
    "preparing": "DPA",
    "run-validation": "Computing",
    "check-validation": "DPA",
    "ready": "Computing",
    "running-concurrent": "DPA",
    "update-end-run": "Computing",
    "running": "Computing",
    "debugging": "DPA",
    "checking": "DPA",
    "done": None,
}

STREAM_TO_EVENTTYPE = {
    "ION": 90700000,
    "IONRAW": 90800000,
    "TURBORAW": 90400000,
    "BEAMGAS": 97000000,
    "FULL": 90000000,
    "TURBO": 94000000,
    "HLT2CALIB": 95200000,
    "TURCAL": 95100000,
    "NOBIAS": 96000000,
    "PASSTHROUGH": 98100000,
    "LUMI": 93000000,
    "SMOGPHY": 90300000,
    "LOWMULT": 90600000,
    "EXPRESS": 91000000,
    "CALIB": 95000000,
    "ERRORS": 92000000,
    "HLTONE": 98000000,
    "L0YES": 99000000,
}
EVENTTYPE_TO_STREAM = {v: k for k, v in STREAM_TO_EVENTTYPE.items()}
if len(STREAM_TO_EVENTTYPE) != len(EVENTTYPE_TO_STREAM):
    raise ValueError("Duplicate event type")


LOGBOOK_SYSTEM_TAGS = {
    "Accounting": "System_0",
    "Applications": "System_1",
    "Bookkeeping": "System_2",
    "CS": "System_3",
    "CernVM-fs": "System_4",
    "Cloud": "System_5",
    "ConDB": "System_6",
    "Data Management": "System_7",
    "FTS": "System_8",
    "FileCatalog": "System_9",
    "Freezer": "System_10",
    "GGUS": "System_11",
    "HowTo": "System_12",
    "LogSE": "System_13",
    "Monitoring": "System_14",
    "Operations minutes": "System_15",
    "Production": "System_16",
    "RMS": "System_17",
    "Releases": "System_18",
    "Site Downtime": "System_19",
    "Site Management": "System_20",
    "Stager": "System_21",
    "Stripping": "System_22",
    "Turbo": "System_23",
    "User": "System_24",
    "VOBox": "System_25",
    "Web portal": "System_26",
    "WMS": "System_27",
    "DIrac": "System_28",
    "WGP": "System_29",
    "RSS": "System_30",
    "TS": "System_31",
    "Sprucing": "System_32",
}


class BaseModel(_BaseModel):
    class Config:
        extra = "forbid"


class Validation(BaseModel):
    transform_ids: list[int]
    n_files_expected: int | None = None
    running: bool = True
    cleaned: bool = False
    file_status: dict[int, dict[str, int]] = {}


class RequestChecks(BaseModel):
    dm_check: dict[int, bool | None] = {}
    pm_check: dict[int, bool | None] = {}
    dm_clean: dict[int, bool | None] = {}


class Request(BaseModel):
    transform_ids: list[int]
    removal: int | None = None
    replication: int | None = None
    file_status: dict[int, dict[str, int]] = {}
    checks: RequestChecks = RequestChecks()

    @property
    def all_transform_ids(self):
        ids = []
        ids.extend(self.transform_ids)
        if self.removal is not None:
            ids.append(self.removal)
        if self.replication is not None:
            ids.append(self.replication)
        return ids


class RepoIssueMetadata(BaseModel):
    validations: dict[int, Validation] = {}
    requests: dict[int, Request] = {}
    last_updated: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))


class RepoIssue:
    def __init__(self, issue: gitlab.v4.objects.ProjectIssue):
        self.issue = issue
        self.run_yaml_blob, self.request_yaml_blob = self._extract_request(self.issue)
        self.run_yaml = self._parse_run_yaml_blob(self.run_yaml_blob)
        self.request_yaml = self._parse_request_yaml_blob(self.request_yaml_blob)
        try:
            self.metadata: RepoIssueMetadata = self._extract_metadata()
        except Exception:
            print(f"Failed to extract metadata for {self}")
            raise

    def __repr__(self):
        return f"RepoIssue({self.url})"

    def __hash__(self) -> int:
        return hash(self.url)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, RepoIssue):
            return False
        return self.url == o.issue.attributes["web_url"]

    @property
    def state(self):
        matched = [k for k in self.labels if k.startswith("state::")]
        if len(matched) != 1:
            raise NotImplementedError(self, matched)
        return matched[0].split("::")[1]

    @property
    def url(self):
        return self.issue.attributes["web_url"]

    @property
    def labels(self):
        return self.issue.labels

    @property
    def discussions(self):
        return self.issue.discussions

    @staticmethod
    def _extract_request(issue):
        issue_url = issue.attributes["web_url"]
        run_yaml_blob = None
        request_yaml_blob = None
        for lang, blob in re.findall("```([^`\n]*)\n([^`]+)```", issue.description):
            lang = lang.strip().lower()
            if lang and lang != "yaml":
                continue
            try:
                loaded = yaml.safe_load(blob)
            except yaml.YAMLError as e:
                raise NotImplementedError(issue_url) from e
            if isinstance(loaded, dict):
                if run_yaml_blob is not None:
                    raise NotImplementedError(issue_url)
                run_yaml_blob = blob
                continue
            if isinstance(loaded, list):
                if request_yaml_blob is not None:
                    raise NotImplementedError(issue_url)
                request_yaml_blob = blob
                continue
            raise NotImplementedError(issue_url)

        if run_yaml_blob is None or request_yaml_blob is None:
            raise NotImplementedError(issue_url)

        return run_yaml_blob, request_yaml_blob

    def _parse_run_yaml_blob(self, run_yaml_blob):
        run_yaml = yaml.safe_load(run_yaml_blob)
        if set(run_yaml) != {"start_run", "end_run", "concurrent"}:
            raise NotImplementedError()
        if run_yaml["concurrent"]:
            if run_yaml["start_run"] is None or not isinstance(run_yaml["start_run"], int):
                raise NotImplementedError(self)
            if run_yaml["end_run"] is None or not isinstance(run_yaml["end_run"], int):
                raise NotImplementedError(self)
        if not isinstance(run_yaml["concurrent"], bool):
            raise NotImplementedError()
        return run_yaml

    def _parse_request_yaml_blob(self, request_yaml_blob):
        request_yaml = yaml.safe_load(request_yaml_blob)
        if len(request_yaml) != 1:
            raise NotImplementedError(self)
        request_yaml[0].pop("author", None)
        return request_yaml[0]

    def _extract_metadata(self):
        if self.issue.description.count("Prod Metadata") > 1:
            raise NotImplementedError(self.issue.attributes["web_url"])
        if METADATA_HEADER not in self.issue.description:
            self.issue.description += f"\n\n{METADATA_HEADER}\n\n```json\n{{}}\n```\n\n</details>"
        metadata = RE_METADATA.search(self.issue.description).group(2)
        return RepoIssueMetadata.model_validate_json(metadata)

    def update_metadata(self):
        original = self._extract_metadata().model_dump()
        original.pop("last_updated", None)
        new = self.metadata.model_dump()
        new.pop("last_updated", None)
        if original == new:
            return
        self.issue.description, n = RE_METADATA.subn(
            rf"\1{self.metadata.model_dump_json(indent=4)}\3",
            self.issue.description,
        )
        if n != 1:
            raise NotImplementedError(self)
        self.issue.save()


class DeviceFlowCredentials(metaclass=ABCMeta):
    sso_name: str
    use_pkce: bool

    def __init__(self, client_id, *, client_secret=None, scope=None, token_endpoint=None, device_flow_endpoint=None):
        self.client_id = client_id
        self.client_secret = client_secret
        self.token_endpoint = token_endpoint
        self.device_flow_endpoint = device_flow_endpoint
        self.scope = scope
        self.refresh_token = None
        if self.cache_path.exists():
            cache = json.loads(self.cache_path.read_text())
            self.refresh_token = cache["refresh_token"]
            self.refresh_token_expires_at = cache["refresh_token_expires_at"]
        self.do_token_refresh()

    @property
    def cache_path(self):
        if not CACHE_DIR.exists():
            CACHE_DIR.mkdir(parents=True, exist_ok=True)
        return CACHE_DIR / f"{self.client_id}.json"

    @property
    def access_token(self):
        if self.access_token_expires_at < time.time() + 60:
            self.do_token_refresh()
        return self._access_token

    def update_access_token(self, token_response):
        self._access_token = token_response["access_token"]
        self.access_token_expires_at = time.time() + token_response["expires_in"]

    def update_refresh_token(self, token_response):
        self.refresh_token = token_response["refresh_token"]
        self.refresh_token_expires_at = None
        if "refresh_token_expires_in" in token_response:
            self.refresh_token_expires_at = time.time() + token_response["refresh_token_expires_in"]
        self.cache_path.parent.mkdir(parents=True, exist_ok=True)
        self.cache_path.write_text(
            json.dumps(
                {
                    "refresh_token": self.refresh_token,
                    "refresh_token_expires_at": self.refresh_token_expires_at,
                }
            )
        )

    def do_token_refresh(self):
        if self.refresh_token is None:
            self.device_authorization_login()
            return
        if self.refresh_token_expires_at and self.refresh_token_expires_at < time.time() + 60:
            self.device_authorization_login()
            return
        data = {
            "client_id": self.client_id,
            "grant_type": "refresh_token",
            "refresh_token": self.refresh_token,
        }
        if self.client_secret:
            data["client_secret"] = self.client_secret
        r = requests.post(self.token_endpoint, data=data, timeout=30)
        if not r.ok:
            print(f"Failed to refresh token, with {r.text}")
            self.device_authorization_login()
            return
        token_response = r.json()
        if token_response.get("refresh_token", self.refresh_token) != self.refresh_token:
            self.update_refresh_token(token_response)
        self.update_access_token(token_response)

    def device_authorization_login(self):
        """Get an OIDC token by using Device Authorization Grant"""
        data = {"client_id": self.client_id}
        if self.scope:
            data["scope"] = self.scope
        if self.use_pkce:
            random_state = binascii.hexlify(os.urandom(8))
            code_verifier = binascii.hexlify(os.urandom(48))
            code_challenge = base64.urlsafe_b64encode(hashlib.sha256(code_verifier).digest()).decode().replace("=", "")
            data["state"] = random_state
            data["code_challenge_method"] = "S256"
            data["code_challenge"] = code_challenge
        r = requests.post(self.device_flow_endpoint, data=data, timeout=30)
        if not r.ok:
            raise NotImplementedError(r.text)

        auth_response = r.json()

        print(f"{self.sso_name}\n")
        print("Open the following link directly and follow the instructions:")
        if "verification_uri_complete" in auth_response:
            print(f"    {auth_response['verification_uri_complete']}\n")
        else:
            print(f"    {auth_response['verification_url']}\n")
            print(f"    Code: {auth_response['user_code']}\n")
        print("Waiting for login...")

        while True:
            time.sleep(auth_response.get("interval", 5))
            data = {
                "client_id": self.client_id,
                "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                "device_code": auth_response["device_code"],
            }
            if self.client_secret:
                data["client_secret"] = self.client_secret
            if self.use_pkce:
                data["code_verifier"] = code_verifier
            r = requests.post(self.token_endpoint, data=data, timeout=30)
            if r.ok:
                token_response = r.json()
                self.update_access_token(token_response)
                self.update_refresh_token(token_response)
                break


class CernSSOCredentials(DeviceFlowCredentials):
    sso_name = "CERN Single Sign-On"
    use_pkce = True

    def __init__(self, client_id):
        super().__init__(
            client_id,
            token_endpoint="https://auth.cern.ch/auth/realms/cern/protocol/openid-connect/token",
            device_flow_endpoint="https://auth.cern.ch/auth/realms/cern/protocol/openid-connect/auth/device",
        )


class OperationsLogbook:
    url = "https://lblogbook.cern.ch/Operations/"

    def __init__(self):
        self.credentials = CernSSOCredentials("lhcb-logbook-run3")
        self._username = None
        self._password = None
        self._cookies = None
        if self.cache_path.exists():
            cache = json.loads(self.cache_path.read_text())
            self._cookies = cache["cookies"]
            self._expires = cache["expires"]
        if not self.check_credentials():
            self.login()

    def __str__(self):
        return f"OperationsLogbook({self.url}, username={self._username})"

    @property
    def cache_path(self):
        if not CACHE_DIR.exists():
            CACHE_DIR.mkdir(parents=True, exist_ok=True)
        return CACHE_DIR / "operations-logbook-session.txt"

    @property
    def headers(self):
        return {"Authorization": f"Bearer {self.credentials.access_token}"}

    @property
    def cookies(self):
        if self._cookies is None or self._expires < time.time() + 60:
            self.login()
        return self._cookies

    @cookies.setter
    def cookies(self, cookies):
        self._cookies = cookies.get_dict()
        self._expires = min(c.expires for c in cookies)
        self.cache_path.parent.mkdir(parents=True, exist_ok=True)
        self.cache_path.write_text(
            json.dumps(
                {
                    "cookies": self._cookies,
                    "expires": self._expires,
                }
            )
        )

    def login(self):
        while True:
            if self._username is None:
                self._username = prompt_toolkit.prompt("Enter Logbook Username: ")
            if self._password is None:
                self._password = prompt_toolkit.prompt("Enter Logbook Password: ", is_password=True)
            data = {
                "redir": "",
                "uname": self._username,
                "upassword": self._password,
                "remember": "1",
            }
            r = requests.post(
                self.url, files={k: (None, v) for k, v in data.items()}, headers=self.headers, allow_redirects=False
            )
            if r.status_code == 302:
                self.cookies = r.cookies
                if self.check_credentials():
                    break
            print("Login failed, please try again.")
            self._username = None
            self._password = None

    def check_credentials(self):
        try:
            user = re.search(r'Logged in as "([^"]+)"', self._get(self.url).text).group(1)
        except Exception as e:
            print(f"Failed to check credentials: {e}")
            return False
        else:
            self._username = user
            return True

    @staticmethod
    def _extract_form_data(r):
        soup = BeautifulSoup(r.text, "html.parser")
        form = soup.find("form")
        if not form:
            raise NotImplementedError()
        default_form_data = {}
        # Find all input fields in the form
        for input_tag in form.find_all("input"):
            name = input_tag.get("name")
            value = input_tag.get("value", "")
            if name:
                default_form_data[name] = value

        form_data = {
            "unm": (None, default_form_data["unm"]),
            "upwd": (None, default_form_data["upwd"]),
            "jcmd": (None, default_form_data["jcmd"]),
            "smcmd": (None, default_form_data["smcmd"]),
            "inlineatt": (None, default_form_data["inlineatt"]),
            "new_entry": (None, default_form_data["new_entry"]),
            "entry_modified": (None, default_form_data["entry_modified"]),
            "entry_date": (None, default_form_data["entry_date"]),
            "Author": (None, default_form_data["Author"]),
            "Production_number": (None, default_form_data["Production_number"]),
            "GGUS_Ticket": (None, default_form_data["GGUS_Ticket"]),
            "Trello_JIRA_ticket": (None, default_form_data["Trello_JIRA_ticket"]),
            "CC": (None, default_form_data["CC"]),
            "Subject": (None, default_form_data["Subject"]),
            "Modified": (None, default_form_data["Modified"]),
            "Prefix": (None, default_form_data["Prefix"]),
            "Text": (None, default_form_data.get("Text", "")),
            "encoding": (None, default_form_data["encoding"]),
            "next_attachment": (None, default_form_data["next_attachment"]),
        }
        for i in range(1, int(default_form_data["next_attachment"])):
            form_data[f"attachment{i-1}"] = (None, default_form_data[f"attachment{i-1}"])
        form_data["attfile"] = (None, default_form_data["attfile"])
        return form_data

    def _get(self, url, **kwargs):
        r = requests.get(url, headers=self.headers, cookies=self.cookies, timeout=30, **kwargs)
        r.raise_for_status()
        return r

    def _post(self, url, **kwargs):
        r = requests.post(url, headers=self.headers, cookies=self.cookies, timeout=30, **kwargs)
        r.raise_for_status()
        return r

    def create_post(self, issue_url, transformation_ids, subject, body, attachments, systems: list[str]):
        n = 1
        while True:
            try:
                self._create_post(issue_url, transformation_ids, subject, body, attachments, systems)
            except Exception as e:
                print(f"Failed to create logbook entry (attempt {n}): {e}")
                answer = Prompt.ask("Would you like to retry?", choices=["y", "n"])
                if answer.lower() != "y":
                    return
                n += 1
            else:
                break

    def _create_post(self, issue_url, transformation_ids, subject, body, attachments, systems):
        print(f"Creating logbook entry for {issue_url}")
        print(f"Transformations: {transformation_ids}")
        print(f"Subject: {subject}")
        print(f"Body:\n{body}")
        print(f"Attachments: {list(attachments)}")
        print(f"Systems: {systems}")
        form_data = self._extract_form_data(self._get(f"{self.url}?cmd=New"))

        for name, blob in attachments.items():
            form_data["attfile"] = (name, blob)
            form_data["cmd"] = (None, "Upload")
            form_data = self._extract_form_data(self._post(self.url, files=form_data))

        form_data["Subject"] = (None, subject)
        form_data["Production_number"] = (None, ",".join(map(str, transformation_ids)))
        form_data["Trello_JIRA_ticket"] = (None, issue_url)
        form_data["Text"] = (None, body)
        for system in systems:
            form_data[LOGBOOK_SYSTEM_TAGS[system]] = (None, system)
        form_data["cmd"] = (None, "Submit")
        r = self._post(self.url, files=form_data, allow_redirects=False)
        print(f"Logbook entry created at {r.headers['Location']}")


class ProdRequestsGitlabRepo:
    def __init__(self, *, with_auth: bool):
        cred_path = CACHE_DIR / "gitlab.json"
        cred = {}
        if cred_path.exists():
            cred = json.loads(cred_path.read_text())
        if with_auth:
            while True:
                private_token = cred.get("private_token")
                self.api = gitlab.Gitlab(url="https://gitlab.cern.ch", private_token=private_token)
                try:
                    self.api.auth()
                except gitlab.exceptions.GitlabAuthenticationError:
                    print("Gitlab authentication failed, please provide a new token.")
                    cred = {
                        "private_token": prompt_toolkit.prompt(
                            "Go to https://gitlab.cern.ch/lhcb-dpa/prod-requests/-/settings/access_tokens"
                            " and create a new token with':\n    * role; Developer\n"
                            "    * scopes: api\nThen paste it here:\n",
                            is_password=True,
                        )
                    }
                    cred_path.write_text(json.dumps(cred))
                else:
                    break
        else:
            self.api = gitlab.Gitlab(url="https://gitlab.cern.ch")

        self.project = self.api.projects.get("lhcb-dpa/prod-requests")

    @property
    def known_labels(self):
        if not hasattr(self, "_known_labels"):
            self._known_labels = {label.name: label.color for label in self.project.labels.list(iterator=True)}
        return self._known_labels

    @known_labels.deleter
    def known_labels(self):
        del self._known_labels

    def __str__(self):
        return f"ProdRequestsGitlabRepo({self.project.attributes['web_url']})"

    def get_issue(self, issue: int | str):
        if isinstance(issue, str):
            if not issue.startswith(self.project.attributes["web_url"]):
                raise ValueError("Invalid issue URL")
            issue = int(issue.split("/")[-1].split("#")[0])
        return self.project.issues.get(issue)

    def poll(self, *, states=ISSUE_STATE_RESPONSIBLE, do_status_update=False):
        all_tids = set()
        results = defaultdict(list)

        with Progress(*PROGRESS_COLUMNS) as progress:
            task = progress.add_task("Polling issues...", total=len(states))
            for state in states:
                issues = self.project.issues.list(labels=f"state::{state}", state="opened", get_all=True)
                task2 = progress.add_task(f"Polling {state} issues...", total=len(issues))
                for issue in issues:
                    issue = RepoIssue(issue)
                    results[state].append(issue)
                    for request_meta in issue.metadata.validations.values():
                        all_tids |= set(request_meta.transform_ids)
                    for request_meta in issue.metadata.requests.values():
                        all_tids |= set(request_meta.all_transform_ids)
                    progress.advance(task2)
                progress.remove_task(task2)
                progress.update(task, advance=1)
        # If replication or removal has not been submitted yet we might have a None in the set
        if None in all_tids:
            all_tids.remove(None)

        if do_status_update:
            now = datetime.now(timezone.utc)
            file_statuses = asyncio.run(get_file_statuses(all_tids))
            with Progress(*PROGRESS_COLUMNS) as progress:
                task = progress.add_task("Updating issue metadata...", total=sum(map(len, results.values())))
                for state, issues in results.items():
                    for issue in issues:
                        for validation_meta in issue.metadata.validations.values():
                            for id in validation_meta.transform_ids:
                                if id in file_statuses:
                                    validation_meta.file_status[id] = file_statuses[id]
                        for request_meta in issue.metadata.requests.values():
                            for id in request_meta.all_transform_ids:
                                if id in file_statuses:
                                    request_meta.file_status[id] = file_statuses[id]
                        issue.metadata.last_updated = now
                        issue.update_metadata()
                        progress.update(task, advance=1)

        return dict(results)


async def get_file_status(tid):
    loop = asyncio.get_running_loop()
    return returnValueOrRaise(
        await loop.run_in_executor(
            None,
            TransformationClient().getCounters,
            "TransformationFiles",
            ["TransformationID", "Status"],
            {"TransformationID": tid},
        )
    )


async def get_file_statuses(tids):
    tids = {int(tid) for tid in tids}
    file_statuses = defaultdict(dict)
    with Progress(*PROGRESS_COLUMNS) as progress:
        task1 = progress.add_task("Getting file counts...", total=len(tids))
        for result in asyncio.as_completed(map(get_file_status, tids)):
            for meta, count in await result:
                file_statuses[int(meta["TransformationID"])][meta["Status"]] = count
            progress.update(task1, advance=1)
    return file_statuses



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Insert a new set of simulation conditions in the Bookkeeping.

To run this script interactively pass no keyword arguments. The script will
then prompt you for the required information.

Alternatively, you can run the script in batch mode by --batch. In this case,
you need to provide all the required information as keyword arguments.
"""
from functools import partial

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script
from DIRAC.Core.Utilities.ReturnValues import convertToReturnValue

REQUIRED_KEYS = {
    "SimDescription",
    "BeamCond",
    "BeamEnergy",
    "Generator",
    "MagneticField",
    "DetectorCond",
    "Luminosity",
    "G4settings",
}


def parseArgs():
    condDict = {"Visible": "Y"}
    batch = False

    @convertToReturnValue
    def setKey(key: str, value: str):
        condDict[key] = value

    @convertToReturnValue
    def setBatch(_):
        nonlocal batch
        batch = True

    switches = [
        ("", "sim-description=", "SimDescription", partial(setKey, "SimDescription")),
        ("", "beam-cond=", "BeamCond", partial(setKey, "BeamCond")),
        ("", "beam-energy=", "BeamEnergy", partial(setKey, "BeamEnergy")),
        ("", "generator=", "Generator", partial(setKey, "Generator")),
        ("", "magnetic-field=", "MagneticField", partial(setKey, "MagneticField")),
        ("", "detector-cond=", "DetectorCond", partial(setKey, "DetectorCond")),
        ("", "luminosity=", "Luminosity", partial(setKey, "Luminosity")),
        ("", "g4settings=", "G4settings", partial(setKey, "G4settings")),
        ("", "batch", "Don't ask for confirmation before inserting", setBatch),
    ]
    Script.registerSwitches(switches)
    Script.parseCommandLine(ignoreErrors=False)

    if batch:
        misingKeys = REQUIRED_KEYS - set(condDict.keys())
        if misingKeys:
            gLogger.error(f"Running with --batch but missing required keys: {misingKeys}")
            DIRAC.exit(1)

    return condDict, batch


@Script()
def main():
    condDict, batch = parseArgs()

    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    bk = BookkeepingClient()

    for key in REQUIRED_KEYS:
        if key not in condDict:
            condDict[key] = input(f"{key}: ")

    if not batch:
        gLogger.notice("Do you want to add these new simulation conditions? (yes or no)")
        value = input("Choice:")
        choice = value.lower()
        if choice not in ["yes", "y"]:
            gLogger.notice("Aborted!")
            DIRAC.exit(2)

    res = bk.insertSimConditions(condDict)
    if not res["OK"]:
        gLogger.error(res["Message"])
        DIRAC.exit(3)

    gLogger.notice("The simulation conditions added successfully!")


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""OutputDataPolicy generates the output data that will be created by a
workflow task.

DIRAC assumes an execute() method will exist during usage.
"""
from DIRAC import gLogger
from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob
from LHCbDIRAC.Core.Utilities.ProductionData import preSubmissionLFNs


class OutputDataPolicy:
    """class to generate the output Data."""

    def __init__(self, paramDict):
        """Constructor."""
        self.paramDict = paramDict

    def execute(self):
        """main loop."""
        jobDescription = self.paramDict["Job"]
        prodID = self.paramDict["TransformationID"]
        jobID = self.paramDict["TaskID"]

        job = LHCbJob(jobDescription)
        result = preSubmissionLFNs(job._getParameters(), job.workflow.createCode(), productionID=prodID, jobID=jobID)
        if not result["OK"]:
            gLogger.error(result)
        return result



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Analyse XMLSummary module."""
import os

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Resources.Catalog.PoolXMLFile import getGUID
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from LHCbDIRAC.Core.Utilities.ProductionData import constructProductionLFNs
from LHCbDIRAC.Core.Utilities.XMLSummaries import XMLSummary


class AnalyseXMLSummary(ModuleBase):
    """Analysing the XML summary."""

    def __init__(self, bkClient=None, dm=None):
        """Module initialization."""

        self.log = gLogger.getSubLogger("AnalyseXMLSummary")
        super().__init__(self.log, bkClientIn=bkClient, dm=dm)

        self.nc = NotificationClient()
        self.XMLSummary = ""
        self.XMLSummary_o = None

    def _resolveInputVariables(self):
        """By convention any workflow parameters are resolved here."""

        super()._resolveInputVariables()
        super()._resolveInputStep()

        self.XMLSummary_o = XMLSummary(self.XMLSummary, log=self.log)

    def execute(
        self,
        production_id=None,
        prod_job_id=None,
        wms_job_id=None,
        workflowStatus=None,
        stepStatus=None,
        wf_commons=None,
        step_commons=None,
        step_number=None,
        step_id=None,
    ):
        """Main execution method.

        Here we analyse what is written in the XML summary, and take
        decisions accordingly
        """

        try:
            super().execute(
                production_id,
                prod_job_id,
                wms_job_id,
                workflowStatus,
                stepStatus,
                wf_commons,
                step_commons,
                step_number,
                step_id,
            )

            self._resolveInputVariables()

            self.log.info(f"Performing XML summary analysis for {self.XMLSummary}")
            # Resolve the step and job input data

            self.step_commons["XMLSummary_o"] = self.XMLSummary_o

            failTheJob = False
            if (
                self.XMLSummary_o.success == "True"
                and self.XMLSummary_o.step == "finalize"
                and self.XMLSummary_o._outputsOK()
                and not self.XMLSummary_o.inputFileStats["mult"]
                and not self.XMLSummary_o.inputFileStats["other"]
            ):
                # basic success, now check for failures in the input files
                failTheJob = self._basicSuccess()
            else:
                # here fails!
                failTheJob = True

            if failTheJob:
                self._finalizeWithErrors("XMLSummary reports error")

                self.setApplicationStatus("XMLSummary reports error")
                return S_ERROR("XMLSummary reports error")

            # if the XMLSummary looks ok but the step already failed, preserve the previous error
            if not self.stepStatus["OK"]:
                return S_OK()

            self.log.info(f"XML summary {self.XMLSummary}")
            self.setApplicationStatus(f"{self.applicationName} Step OK")
            return S_OK()

        except Exception as e:  # pylint:disable=broad-except
            self.log.exception("Failure in AnalyseXMLSummary execute module", lException=e)
            self.setApplicationStatus(repr(e))
            return S_ERROR(str(e))

        finally:
            super().finalize()

    def _basicSuccess(self):
        """Treats basic success, meaning the outputs and the status of the XML
        summary are ok.

        Now, we have to check the input files if they are in "part" or
        "fail"
        """
        failTheJob = False
        if self.XMLSummary_o.inputFileStats["part"]:
            if self.numberOfEvents != -1:
                self.log.info("Input on part is ok, since we are not processing all")
                # this is not an error
            else:
                # report to FileReport
                filesInPart = [x[0].strip("LFN:") for x in self.XMLSummary_o.inputStatus if x[1] == "part"]
                self.log.error(f"Files {';'.join(filesInPart)} are in status 'part'")
                for fileInPart in filesInPart:
                    if fileInPart in self.inputDataList:
                        self.log.error(f"Reporting {fileInPart} as 'Problematic'")
                        self.fileReport.setFileStatus(int(self.production_id), fileInPart, "Problematic")
                        failTheJob = True

        if self.XMLSummary_o.inputFileStats["fail"]:
            # report to FileReport
            filesInFail = [x[0].strip("LFN:") for x in self.XMLSummary_o.inputStatus if x[1] == "fail"]
            self.log.error(f"Files {';'.join(filesInFail)} are in status 'fail'")
            for fileInFail in filesInFail:
                if fileInFail in self.inputDataList:
                    self.log.error(f"Reporting {fileInFail} as 'Problematic'")
                    self.fileReport.setFileStatus(int(self.production_id), fileInFail, "Problematic")
                    failTheJob = True

        return failTheJob

    def _finalizeWithErrors(self, subj):
        """Method that sends an email and uploads intermediate job outputs."""
        # Have to check that the output list is defined in the workflow commons, this is
        # done by the first BK report module that executes at the end of a step but in
        # this case the current step 'listoutput' must be added.
        if "outputList" in self.workflow_commons:
            for outputItem in self.step_commons["listoutput"]:
                if outputItem not in self.workflow_commons["outputList"]:
                    self.workflow_commons["outputList"].append(outputItem)
        else:
            self.workflow_commons["outputList"] = self.step_commons["listoutput"]

        result = constructProductionLFNs(self.workflow_commons, self.bkClient)

        if not result["OK"]:
            self.log.error(f"Could not create production LFNs with message \"{result['Message']}\"")
            raise Exception(result["Message"])

        if "DebugLFNs" not in result["Value"]:
            self.log.error(f"No debug LFNs found after creating production LFNs, result was:{result}")
            raise Exception("DebugLFNs Not Found")

        debugLFNs = result["Value"]["DebugLFNs"]

        subject = (
            "["
            + self.siteName
            + "]["
            + self.applicationName
            + "] "
            + self.applicationVersion
            + ": "
            + subj
            + " "
            + self.production_id
            + "_"
            + self.prod_job_id
            + " JobID="
            + str(self.jobID)
        )
        msg = "The Application " + self.applicationName + " " + self.applicationVersion + " had a problem \n"
        msg = msg + "at site " + self.siteName + "\n"
        msg = msg + "JobID is " + str(self.jobID) + "\n"
        msg = msg + "JobName is " + self.production_id + "_" + self.prod_job_id + "\n"

        toUpload = {}
        for lfn in debugLFNs:
            if os.path.exists(os.path.basename(lfn)):
                toUpload[os.path.basename(lfn)] = lfn

        if toUpload:
            msg += "\n\nIntermediate job data files:\n"

        for fname, lfn in toUpload.items():
            guidResult = getGUID(fname)

            guidInput = ""
            if not guidResult["OK"]:
                self.log.error(f"Could not find GUID for {fname} with message", guidResult["Message"])
            elif guidResult["generated"]:
                self.log.info(
                    "PoolXMLFile generated GUID(s) for the following files ", ", ".join(guidResult["generated"])
                )
                guidInput = guidResult["Value"][fname]
            else:
                guidInput = guidResult["Value"][fname]

            if self._WMSJob():
                self.log.info(
                    'Attempting: dm.putAndRegister("%s","%s","%s","%s") on master catalog'
                    % (fname, lfn, guidInput, self.debugSE)
                )
                result = DataManager(masterCatalogOnly=True).putAndRegister(lfn, fname, self.debugSE, guidInput)
                self.log.info(result)
                if not result["OK"]:
                    self.log.error("Could not save INPUT data file with result", str(result["Message"]))
                    msg += f"Could not save intermediate data file {fname} with result\n{result['Message']}\n"
                else:
                    msg = msg + lfn + "\n" + str(result) + "\n"
            else:
                self.log.info(
                    "JOBID is null, would have attempted to upload: LFN:%s, file %s, GUID %s to %s"
                    % (lfn, fname, guidInput, self.debugSE)
                )

        if not self._WMSJob():
            self.log.info(
                f"JOBID is null, *NOT* sending mail, for information the mail was:\n====>Start\n{msg}\n<====End"
            )
        else:
            mailAddress = self.opsH.getValue("EMail/JobFailures", "lhcb-datacrash@cern.ch")
            self.log.info(f"Sending crash mail for job to {mailAddress}")

            res = self.nc.sendMail(mailAddress, subject, msg, "joel.closier@cern.ch", localAttempt=False)
            if not res["OK"]:
                self.log.warn("The mail could not be sent")



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: SEUsageAgent

.. module: SEUsageAgent

:synopsis: SEUsageAgent browses the SEs to determine their content and store it into a DB.
"""
# # imports
import os
import time
import tarfile
import signal
from datetime import datetime
from urllib.request import urlopen
from urllib.error import HTTPError

# # from DIRAC
from DIRAC import S_OK, S_ERROR, rootPath, gConfig
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Utilities.SiteSEMapping import getSEsForSite
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.Interfaces.API.Dirac import Dirac

# # from LHCbDIRAC
from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient
from LHCbDIRAC.DataManagementSystem.DB.StorageUsageDB import StorageUsageDB

AGENT_NAME = "DataManagement/SEUsageAgent"


def alarmTimeoutHandler(*args):
    """handler for signal.SIGALRM."""
    raise Exception("Timeout")


class SEUsageAgent(AgentModule):
    """..class:: SEUsageAgent

    :param mixed storageUsage: StorageUsageDB or StorageUsage client
    :param DataManager dataManager: DataManager instance
    :param Operations opHelper: operations helper
    :param int maximumDelaySinceSD: max delay after SE dump creation
    :param list activeSites: active sites
    """

    def __init__(self, *args, **kwargs):
        """c'tor."""
        AgentModule.__init__(self, *args, **kwargs)

        self.storageUsage = None
        self.dataManager = None
        self.opHelper = None
        self.diracApi = None

        self.activeSites = None

        self.specialReplicas = []
        self.workDirectory = ""
        # maximum delay after storage dump creation
        self.maximumDelaySinceSD = 43200
        self.spaceTokToIgnore = None  # STs to skip during checks
        self.pathToUploadResults = "/lhcb/test/dataNotRegistered"
        self.inputFilesLocation = ""
        self.spaceTokens = {}
        self.siteConfig = {}

    def initialize(self):
        """agent initialisation."""
        # This sets the Default Proxy to used as that defined under
        # /Operations/Shifter/DataManager
        # the shifterProxy option in the Configsorteduration can be used to change this default.
        self.am_setOption("shifterProxy", "DataManager")
        self.am_setModuleParam("shifterProxyLocation", f"{rootPath}/runit/{AGENT_NAME}/proxy")

        if self.am_getOption("DirectDB", False):
            self.storageUsage = StorageUsageDB()
        else:
            self.storageUsage = StorageUsageClient()
        # # replica mgr
        self.dataManager = DataManager()
        # # operations helper
        self.opHelper = Operations()
        # # Dirac API
        self.diracApi = Dirac()
        self.activeSites = self.am_getOption("ActiveSites", None)

        self.specialReplicas = self.am_getOption("SpecialReplicas", [])
        self.workDirectory = self.am_getOption("WorkDirectory", None)
        # maximum delay after storage dump creation
        self.maximumDelaySinceSD = self.am_getOption("MaximumDelaySinceStorageDump", 43200)
        self.spaceTokToIgnore = self.am_getOption("SpaceTokenToIgnore", None)  # STs to skip during checks
        self.pathToUploadResults = self.am_getOption("PathToUploadResults", "/lhcb/test/dataNotRegistered")

        mkDir(self.workDirectory)
        self.log.info(f"Working directory is {self.workDirectory}")
        self.inputFilesLocation = self.am_getOption("InputFilesLocation", "")
        mkDir(self.inputFilesLocation)
        self.log.info(f"Input files will be downloaded to: {self.inputFilesLocation}")
        # FIXME: totally useless
        if self.specialReplicas:
            self.log.info(f"Special replica types: {self.specialReplicas}")
        else:
            self.log.info("No special replica configured.")
        self.log.info(f"Path to upload results: {self.pathToUploadResults} ")
        return S_OK()

    def execute(self):
        """execution in one cycle.

        Loops on the input files to read the content of Storage Elements,
        process them, and store the result into the DB. It reads directory
        by directory (every row of the input file being a directory). If the
        directory exists in the StorageUsage su_Directory table, and if a
        replica also exists for the given SE in the su_SEUsage table, then
        the directory and its usage are stored in the replica table (the
        se_Usage table) together with the insertion time, otherwise it is
        added to the problematic data table (problematicDirs)
        """

        self.log.info("Starting the execute method")
        self.log.info(f"Sites active for checks: {self.activeSites} ")
        siteList = self.activeSites.split(",")
        timingPerSite = {}
        self.spaceTokens = {}
        self.siteConfig = {}
        siteList.sort()
        for lcgSite in siteList:
            site = lcgSite.split(".")[1]
            timingPerSite[site] = {}
            start = time.time()
            res = self.setupSiteConfig(lcgSite)
            if not res["OK"]:
                self.log.warn(f"Error during site configuration {res['Message']} ")
                continue

            self.log.info(f"Parse and read input files for site: {site}")
            res = self.readInputFile(site)
            if not res["OK"]:
                self.log.error(f"Failed to read input files for site {site} ")
                continue
            if res["Value"] < 0:
                self.log.warn(f"Input files read successfully, but not valid for checks (too old?). Skip site {site} ")
                continue

            # ONLY FOR DEBUGGING
            debug = False
            if debug:  # get the problematic summary at the beginning (to debug this part without waiting until the end)
                res = self.getProblematicDirsSummary(site)
                if not res["OK"]:
                    return S_ERROR(res["Message"])
            else:  # follow the usual work flow
                # Flush the problematicDirs table
                self.log.info(f"Flushing the problematic directories table for site {site}...")
                res = self.storageUsage.removeAllFromProblematicDirs(site)
                if not res["OK"]:
                    self.log.error(
                        "Error for site %s: Could not remove old entries from the problematicDirs table! %s:"
                        % (site, res["Message"])
                    )
                    continue
                self.log.info(
                    "Removed %d entries from the problematic directories table for site %s" % (res["Value"], site)
                )
            # END OF DEBUGGING MODIFICATION ###########

            # Start the main loop:
            # read all file of type directory summary for the given site:
            pathToSummary = self.siteConfig[site]["pathToInputFiles"]
            for fileP3 in os.listdir(pathToSummary):
                fullPathFileP3 = os.path.join(pathToSummary, fileP3)
                if "dirSummary" not in fileP3:
                    continue
                self.log.info(f"Start to scan file {fullPathFileP3}")
                for line in open(fullPathFileP3).readlines():
                    splitLine = line.split()
                    try:
                        spaceToken = splitLine[0]
                        storageDirPath = splitLine[1]
                        files = splitLine[2]
                        size = splitLine[3]
                    except IndexError:
                        self.log.error(f"ERROR in input data format. Line is: {line}")
                        continue
                    oneDirDict = {}
                    dirPath = storageDirPath
                    # specialRep = False
                    replicaType = "normal"
                    for sr in self.specialReplicas:
                        prefix = "/lhcb/" + sr
                        self.log.verbose(f"Trying to match prefix {prefix} ")
                        if prefix in dirPath:
                            # strip the initial prefix, to get the LFN as registered in the FC
                            dirPath = storageDirPath.split(prefix)[1]
                            replicaType = sr
                            self.log.info(f"prefix: {prefix} \n storageDirPath: {storageDirPath}\n dirPath: {dirPath}")
                            break
                    oneDirDict[dirPath] = {
                        "SpaceToken": spaceToken,
                        "Files": files,
                        "Size": size,
                        "Updated": "na",
                        "Site": site,
                        "ReplicaType": replicaType,
                    }
                    # the format of the entry to be processed must be a dictionary with LFN path as key
                    # use this format for consistency with already existing methods of StorageUsageDB
                    # which take in input a dictionary like this
                    self.log.info(f"Processing directory: {oneDirDict}")
                    # initialize the isRegistered flag. Change it according to the checks SE vs FC
                    # possible values of isRegistered flag are:
                    # NotRegisteredInFC: data not registered in FC
                    # RegisteredInFC: data correctly registered in FC
                    # MissingDataFromSE: the directory exists in the FC for that SE, but there is
                    # less data on the SE than what reported in the FC
                    isRegistered = False
                    lfcFiles = -1
                    lfcSize = -1
                    self.log.verbose(f"check if dirName exists in su_Directory: {dirPath}")
                    # select entries with that LFN path (no reference to a particular site in this query)
                    res = self.storageUsage.getIDs(oneDirDict)
                    if not res["OK"]:
                        self.log.warn(f"failed to get DID from StorageUsageDB.su_Directory table for dir: {dirPath} ")
                        continue
                    elif not res["Value"]:
                        self.log.info(
                            "NO LFN registered in the FC for the given path %s => insert the entry in the "
                            "problematicDirs table and delete this entry from the replicas table" % dirPath
                        )
                        isRegistered = "NotRegisteredInFC"
                    else:
                        value = res["Value"]
                        self.log.info(f"directory LFN is registered in the FC, output of getIDs is {value}")
                        for directory in value:
                            self.log.verbose("Path: %s su_Directory.DID %d" % (directory, value[directory]))
                        self.log.verbose("check if this particular replica is registered in the FC.")
                        res = self.storageUsage.getAllReplicasInFC(dirPath)
                        if not res["OK"]:
                            self.log.error(f"failed to get replicas for {dirPath} directory ")
                            continue
                        elif not res["Value"]:
                            self.log.warn(
                                "NO replica found for %s on any SE! Anomalous case: the LFN is registered in the FC"
                                "but with NO replica! For the time being, insert it into "
                                "problematicDirs table " % dirPath
                            )
                            # TODO: we should decide what to do in this case.
                            # This might happen, but it is a problem at FC level...
                            isRegistered = "NotRegisteredInFC"
                        else:  # got some replicas! let's see if there is one for this SE
                            associatedDiracSEs = self.spaceTokens[site][spaceToken]["DiracSEs"]
                            self.log.verbose(f"SpaceToken: {spaceToken} list of its DiracSEs: {associatedDiracSEs}")
                            lfcFiles = 0
                            lfcSize = 0
                            for lfn in res["Value"]:
                                matchedSE = ""
                                for se in res["Value"][lfn]:
                                    self.log.verbose(f"SpaceToken: {spaceToken} -- se: {se}")
                                    if se in associatedDiracSEs:
                                        # consider only the FC replicas of the corresponding Dirac SE
                                        if oneDirDict[dirPath]["ReplicaType"] in self.specialReplicas:
                                            self.log.info(f"SpecialReplica: {oneDirDict[dirPath]['ReplicaType']}")
                                            seSuffix = oneDirDict[dirPath]["ReplicaType"].upper()
                                            if seSuffix not in se:
                                                self.log.verbose(f"SE does not contain the suffix: {seSuffix}. Skip it")
                                                continue
                                        lfcFiles += int(res["Value"][lfn][se]["Files"])
                                        lfcSize += int(res["Value"][lfn][se]["Size"])
                                        self.log.info(f"==> the replica is registered in the FC with DiracSE= {se}")
                                        # this is because there can be DST and M-DST files in the same directory
                                        if not matchedSE:
                                            matchedSE = se
                                        else:
                                            matchedSE = matchedSE + ".and." + se
                            # check also whether the number of files and total directory size match:
                            seSize = int(oneDirDict[dirPath]["Size"])
                            seFiles = int(oneDirDict[dirPath]["Files"])
                            self.log.info(f"Matched SE: {matchedSE}")
                            oneDirDict[dirPath]["SEName"] = matchedSE
                            self.log.info(
                                "lfcFiles = %d lfcSize = %d seFiles = %d seSize = %d"
                                % (lfcFiles, lfcSize, seFiles, seSize)
                            )
                            if seSize > lfcSize:
                                self.log.info("Data on SE exceeds what reported in FC! some data not registered in FC")
                                isRegistered = "NotRegisteredInFC"
                            elif seSize < lfcSize:
                                self.log.info("Data on FC exceeds what reported by SE dump! missing data from storage")
                                isRegistered = "MissingDataFromSE"
                            elif lfcFiles == seFiles and lfcSize == seSize:
                                self.log.info("Number of files and total size matches with what reported by FC")
                                isRegistered = "RegisteredInFC"
                            else:
                                self.log.info("Unexpected case: seSize = lfcSize but seFiles != lfcFiles")
                                isRegistered = "InconsistentFilesSize"

                    self.log.info(f"isRegistered flag is {isRegistered}")
                    # now insert the entry into the problematicDirs table,
                    # or the replicas table according the isRegistered flag.
                    if not isRegistered:
                        self.log.error(f"the isRegistered flag could not be updated for this directory: {oneDirDict} ")
                        continue
                    if isRegistered in ("NotRegisteredInFC", "MissingDataFromSE", "InconsistentFilesSize"):
                        self.log.info(
                            "problematic directory! Add the entry %s to problematicDirs table. "
                            "Before (if necessary) remove it from se_Usage table" % (dirPath)
                        )
                        res = self.storageUsage.removeDirFromSe_Usage(oneDirDict)
                        if not res["OK"]:
                            self.log.error(f"failed to remove from se_Usage table: {res['Message']}")
                            continue
                        else:
                            removedDirs = res["Value"]
                            if removedDirs:
                                self.log.verbose(f"entry {oneDirDict} correctly removed from se_Usage table")
                            else:
                                self.log.verbose(f"entry {oneDirDict} was NOT in se_Usage table")
                            # update the oneDirDict and insert into problematicDirs
                            oneDirDict[dirPath]["Problem"] = isRegistered
                            oneDirDict[dirPath]["LFCFiles"] = lfcFiles
                            oneDirDict[dirPath]["LFCSize"] = lfcSize
                            res = self.storageUsage.publishToProblematicDirs(oneDirDict)
                            if not res["OK"]:
                                self.log.error(f"failed to publish entry {dirPath} to problematicDirs table")
                            else:
                                self.log.info(f"entry {oneDirDict} correctly published to problematicDirs table")
                    elif isRegistered == "RegisteredInFC":  # publish to se_Usage table!
                        self.log.verbose(
                            "replica %s is registered! remove from problematicDirs (if necessary) and "
                            " publish to se_Usage table" % dirPath
                        )
                        res = self.storageUsage.removeDirFromProblematicDirs(oneDirDict)
                        if not res["OK"]:
                            self.log.error(f"failed to remove from problematicDirs: {oneDirDict}")
                            continue
                        else:
                            removedDirs = res["Value"]
                            if removedDirs:
                                self.log.verbose(f"entry {oneDirDict} correctly removed from problematicDirs")
                            else:
                                self.log.verbose(f"entry {oneDirDict} was NOT in problematicDirs")
                            res = self.storageUsage.publishToSEReplicas(oneDirDict)
                            if not res["OK"]:
                                self.log.warn(f"failed to publish {oneDirDict} entry to se_Usage table")
                            else:
                                self.log.info(f"entry {oneDirDict} correctly published to se_Usage")
                    else:
                        self.log.error(f"Unknown value of isRegistered flag: {isRegistered} ")

                self.log.info(f"Finished loop on file: {fileP3} ")

            self.log.info(f"Finished loop for site: {site} ")
            # query problematicDirs table to get a summary of directories with the flag: NotRegisteredInFC
            self.log.info("Get the summary of problematic directories..")
            res = self.getProblematicDirsSummary(site)
            if not res["OK"]:
                return res
            end = time.time()
            timingPerSite[site]["timePerCycle"] = end - start

        self.log.info("--------- End of cycle ------------------")
        self.log.info("checked sites:")
        for site, siteTiming in timingPerSite.items():
            self.log.info(f"Site: {site} -  total time {siteTiming}")
        return S_OK()

    def setupSiteConfig(self, lcgSite):
        """Setup the configuration for the site."""
        site = lcgSite.split(".")[1]
        self.spaceTokens[site] = {
            "LHCb-Tape": {"year": "2011", "DiracSEs": [site + "-RAW", site + "-RDST", site + "-ARCHIVE"]},
            "LHCb-Disk": {
                "year": "2011",
                "DiracSEs": [site + "-DST", site + "_MC-DST", site + "-FAILOVER", site + "-BUFFER"],
            },
            "LHCb_USER": {"year": "2011", "DiracSEs": [site + "-USER"]},
            "LHCb_RAW": {"year": "2010", "DiracSEs": [site + "-RAW"]},
            "LHCb_RDST": {"year": "2010", "DiracSEs": [site + "-RDST"]},
            "LHCb_DST": {"year": "2010", "DiracSEs": [site + "-DST"]},
            "LHCb_MC_DST": {"year": "2010", "DiracSEs": [site + "_MC-DST"]},
            "LHCb_FAILOVER": {"year": "2010", "DiracSEs": [site + "-FAILOVER"]},
        }
        # for CERN: two more SEs: FREEZER AND HISTO

        for st in self.spaceTokens[site]:
            if st in self.spaceTokToIgnore:
                self.spaceTokens[site][st]["Check"] = False
            else:
                self.spaceTokens[site][st]["Check"] = True

        rootConfigPath = "DataConsistency"
        res = self.opHelper.getOptionsDict(f"{rootConfigPath}/{site}")
        # res = gConfig.getOptionsDict( "%s/%s" % ( rootConfigPath, site ) )
        if not res["OK"]:
            self.log.error(f"could not get configuration for site {site} : {res['Message']} ")
            return S_ERROR(res["Message"])
        storageDumpFileName = res["Value"]["StorageDumpFileName"]
        storageDumpURL = res["Value"]["StorageDumpURL"]
        fileType = res["Value"]["FileType"]
        pathInsideTar = res["Value"]["PathInsideTar"]
        # res = getSEsForSite( site )
        res = getSEsForSite(lcgSite)
        if not res["OK"]:
            self.log.error(f"could not get SEs for site {site} ")
            return S_ERROR(res["Message"])
        ses = res["Value"]
        self.log.info(f"SEs: {ses}")
        seName = site + "-RAW"
        res = gConfig.getOption(f"/Resources/StorageElements/{seName}/AccessProtocol.1/Path")
        if not res["OK"]:
            self.log.error(f"could not get configuration for SE {res['Message']} ")
            return S_ERROR(res["Message"])
        saPath = res["Value"]

        self.siteConfig[site] = {
            "originFileName": storageDumpFileName,
            "pathInsideTar": pathInsideTar,
            "originURL": storageDumpURL,
            "targetPath": os.path.join(self.inputFilesLocation, "downloadedFiles", site),
            "pathToInputFiles": os.path.join(self.inputFilesLocation, "goodFormat", site),
            "DiracName": lcgSite,
            "FileNameType": fileType,
            "SEs": ses,
            "SAPath": saPath,
        }
        self.log.info(f"Configuration for site {site} : {self.siteConfig[site]} ")
        return S_OK()

    def readInputFile(self, site):
        """Download, read and parse input files with SEs content.

        Write down the results to the ASCII files.
        There are 3 phases in the manipulation of input files:
        1. it is directly the format of the DB query output, right after uncompressing the
        tar file provided by the site:
        PFN | size | update (ms)
        2. one row per file, with format:  LFN size update
        3. directory summary files: one row per directory, with format:
        SE DirectoryLFN NumOfFiles TotalSize Update(actually, not used)
        """

        retCode = 0
        originFileName = self.siteConfig[site]["originFileName"]
        originURL = self.siteConfig[site]["originURL"]
        targetPath = self.siteConfig[site]["targetPath"]
        mkDir(targetPath)
        pathToInputFiles = self.siteConfig[site]["pathToInputFiles"]
        mkDir(pathToInputFiles)
        if pathToInputFiles[-1] != "/":
            pathToInputFiles = f"{pathToInputFiles}/"
        self.log.info(f"Reading input files for site {site} ")
        self.log.info(
            "originFileName: %s, originURL: %s, targetPath: %s, pathToInputFiles: %s "
            % (originFileName, originURL, targetPath, pathToInputFiles)
        )

        if targetPath[-1] != "/":
            targetPath = f"{targetPath}/"
        targetPathForDownload = targetPath + self.siteConfig[site]["pathInsideTar"]
        self.log.info(f"Target path to download input files: {targetPathForDownload}")
        # delete all existing files in the target directory if necessary:
        previousFiles = []
        try:
            previousFiles = os.listdir(targetPathForDownload)
            self.log.info(f"In {targetPathForDownload} found these files: {previousFiles}")
        except TypeError:
            pass
        except OSError:
            self.log.info(f"no leftover to remove from {targetPathForDownload}. Proceed downloading input files...")
        if previousFiles:
            self.log.info(f"delete these files: {previousFiles} ")
            for fileName in previousFiles:
                fullPath = os.path.join(targetPathForDownload, fileName)
                self.log.info(f"removing {fullPath}")
                os.remove(fullPath)

        # Download input data made available by the sites. Reuse the code of dirac-install:
        # urlretrieveTimeout, downloadAndExtractTarball
        res = self.downloadAndExtractTarball(originFileName, originURL, targetPath)
        if not res["OK"]:
            return S_ERROR(res["Message"])

        defaultDate = "na"
        # defaultSize = 0

        self.log.info("Check storage dump creation time..")
        self.StorageDumpCreationTime = None
        res = self.checkCreationDate(targetPathForDownload)
        if not res["OK"]:
            self.log.error(f"Failed to check storage dump date for site {site} ")
            return S_ERROR(res)
        if res["Value"] != 0:
            self.log.warn(f"Stale storage dump for site {site} ")
            retCode = -1
            return S_OK(retCode)
        self.log.info("Creation date check is fine!")

        inputFilesListP1 = os.listdir(targetPathForDownload)
        self.log.info(f"List of raw input files in {targetPathForDownload}: {inputFilesListP1} ")

        # delete all leftovers of previous runs from the pathToInputFiles
        try:
            previousParsedFiles = os.listdir(pathToInputFiles)
            self.log.info(f"In {pathToInputFiles} found these files: {previousParsedFiles}")
        except TypeError:
            pass
        except OSError:
            self.log.info(f"no leftover to remove from {pathToInputFiles} . Proceed to parse the input files...")
        if previousParsedFiles:
            self.log.info(f"delete these files: {previousParsedFiles} ")
            for fileName in previousParsedFiles:
                fullPath = os.path.join(pathToInputFiles, fileName)
                self.log.info(f"removing {fullPath}")
                os.remove(fullPath)

        # If necessary, call the special Castor pre-parser which analyse the name server dump and creates
        # one file per space token. The space token attribution is totally fake! it's based on the namespace.
        # For Castor, no other way to do it.
        if site in ["RAL"]:
            self.log.info("Calling special parser for Castor")
            res = self.castorPreParser(site, targetPathForDownload)
            if not res["OK"]:
                self.log.error("Error parsing storage dump")
            else:
                # recompute the list of input files
                inputFilesListP1 = os.listdir(targetPathForDownload)
                self.log.info(f"List of raw input files in {targetPathForDownload}: {inputFilesListP1} ")

        # if necessary, merge the files in order to have one file per space token: LHCb-Tape, LHCb-Disk, LHCb_USER
        # this is necessary in the transition phase while there are still some data on the old space tokens of 2010
        self.log.info("Merge the input files to have one file per space token")
        # Loop on N files relative to the site (one file for each space token)
        # previously open files for writing
        # every input file corresponds to one space token
        # whereas for the output, there is one file per NEW space token, so a merging is done

        outputFileMerged = {}
        for st in self.spaceTokens[site]:
            outputFileMerged[st] = {}
            if self.spaceTokens[site][st]["year"] == "2011":
                self.log.info(f"Preparing output files for space tokens: {st}")
                # merged file:
                fileP2Merged = pathToInputFiles + st + ".Merged.txt"
                outputFileMerged[st]["MergedFileName"] = fileP2Merged
                self.log.info(f"Opening file {fileP2Merged} in w mode")
                fP2Merged = open(fileP2Merged, "w")
                outputFileMerged[st]["pointerToMergedFile"] = fP2Merged
                # directory summary file:
                fileP3DirSummary = pathToInputFiles + st + ".dirSummary.txt"
                outputFileMerged[st]["DirSummaryFileName"] = fileP3DirSummary
                self.log.info(f"Opening file {fileP3DirSummary} in w mode")
                fP3DirSummary = open(fileP3DirSummary, "w")
                outputFileMerged[st]["pointerToDirSummaryFile"] = fP3DirSummary
        if st in ["LHCb-Disk", "LHCb-EOS", "LHCb_USER"]:
            try:
                outputFileMerged[st] = outputFileMerged["LHCb-Disk"]
            except KeyError:
                self.log.error(f"no pointer to file for st={st} ")
        if st == "LHCb-Tape":
            try:
                outputFileMerged[st] = outputFileMerged["LHCb-Tape"]
            except KeyError:
                self.log.error(f"no pointer to file for st={st} ")
        self.log.info("Parsed output files : ")
        for st in outputFileMerged:
            self.log.info(
                "space token: %s -> \n  %s\n %s  "
                % (st, outputFileMerged[st]["MergedFileName"], outputFileMerged[st]["DirSummaryFileName"])
            )

        for inputFileP1 in inputFilesListP1:
            self.log.info(f"+++++ input file: {inputFileP1} ++++++")
            # the expected input file line is: pfn | size | date
            # manipulate the input file to create a directory summary file (one row per directory)
            # Check the validity of the input file name:
            splitFile = inputFileP1.split(".")
            st = splitFile[0]
            if st not in self.spaceTokens[site]:
                self.log.warn(f"Not a  valid space token in the file name: {inputFileP1} ")
                continue

            completeSTId = st
            if len(splitFile) > 3:
                # file with format provided by SARA e.g. LHCb_RAW.31455.INACTIVE.txt
                completeSTId = splitFile[0] + "." + splitFile[1] + "." + splitFile[2]
            fullPathFileP1 = os.path.join(targetPathForDownload, inputFileP1)
            if not os.path.getsize(fullPathFileP1):
                self.log.info(f"{inputFileP1} file has zero size, will be skipped ")
                continue
            if "INACTIVE" in inputFileP1:
                self.log.info(f"WARNING: File {inputFileP1} will be analysed, even if space token is inactive")
            self.log.info(f"processing input file for space token: {st} ")
            fP2 = outputFileMerged[st]["pointerToMergedFile"]
            fileP2 = outputFileMerged[st]["MergedFileName"]
            self.log.info(f"Reading from file {fullPathFileP1}\n and writing to: {fileP2}")
            totalLines = 0  # counts all lines in input
            processedLines = 0  # counts all processed lines
            dirac_dir_lines = 0
            totalSize = 0
            totalFiles = 0
            for line in open(fullPathFileP1).readlines():
                totalLines += 1
                try:
                    splitLine = line.split("|")
                except BaseException:
                    self.log.error(f"Error: impossible to split line : {line}")
                    continue

                providedFilePath = splitLine[0].rstrip()
                if "dirac_directory" in providedFilePath:
                    dirac_dir_lines += 1
                    continue
                # get LFN from PFN if necessary:
                filePath = ""
                if self.siteConfig[site]["FileNameType"] == "LFN":
                    filePath = providedFilePath
                elif self.siteConfig[site]["FileNameType"] == "PFN":
                    res = self.getLFNPath(site, providedFilePath)
                    if not res["OK"]:
                        self.log.error(f"ERROR getLFNPath returned: {res} ")
                        continue
                    filePath = res["Value"]
                self.log.debug(f"filePath: {filePath}")
                if not filePath:
                    self.log.info(f"it was not possible to get the LFN for PFN={filePath}, skip this line")
                    continue

                self.log.debug(f"splitLine: {splitLine} ")
                size = splitLine[1].lstrip()
                totalSize += int(size)
                totalFiles += 1
                updated = splitLine[2].lstrip()
                newLine = filePath + " " + size + " " + updated
                if newLine[-1] != "\n":
                    newLine = f"{newLine}\n"
                fP2.write(f"{newLine}")
                processedLines += 1
            #        except:
            #          # the last line of these files is empty, so it will give this exception
            #          self.log.error( "Error in input line format! Line is: %s" % line )
            #          continue
            fP2.flush()

            self.log.info(f"Cleaning the STSummary table entries for site {site} and space token {completeSTId} ...")
            res = self.storageUsage.removeSTSummary(site, completeSTId)
            if not res["OK"]:
                self.log.error(
                    f"Unable to remove old entries from the STSummary table for site {site}: {res['Message']} "
                )
                continue
            self.log.info("Removed %d entries from the STSummary table for site %s" % (res["Value"], site))

            self.log.info(
                "%s - %s Total size: %d , total files: %d : publish to STSummary"
                % (site, completeSTId, totalSize, totalFiles)
            )
            res = self.storageUsage.publishTose_STSummary(
                site, completeSTId, totalSize, totalFiles, self.StorageDumpCreationTime
            )
            if not res["OK"]:
                self.log.error(f"failed to publish {site} - {st} summary ")
                return S_ERROR(res)
            self.log.info(
                "Total lines: %d , correctly processed: %d, dirac_directory found %d "
                % (totalLines, processedLines, dirac_dir_lines)
            )
        # close output files:
        for st in outputFileMerged:
            p2 = outputFileMerged[st]["pointerToMergedFile"]
            p2.close()

        # produce directory summaries:
        mergedFilesList = os.listdir(pathToInputFiles)
        for fileName in mergedFilesList:
            if "Merged" not in fileName:
                continue
            fileP2 = os.path.join(pathToInputFiles, fileName)
            self.log.info(f"Reading from Merged file fileP2 {fileP2} ")
            for spaceTok in self.spaceTokens[site]:
                self.log.debug(f"..check for space token: {spaceTok}")
                if spaceTok in fileP2:
                    st = spaceTok
                    break

            if not self.spaceTokens[site][st]["Check"]:
                self.log.info(f"Skip this space token: {st}")
                continue
            self.log.info(f"Space token: {st}")
            totalLines = 0  # counts all lines in input
            processedLines = 0  # counts all processed lines
            self.dirDict = {}
            for line in open(fileP2).readlines():
                self.log.debug(f"..processing line: {line}")
                totalLines += 1
                splitLine = line.split()
                filePath = splitLine[0]
                updated = defaultDate
                size = int(splitLine[1])
                # # currently the creation data is NOT stored in the DB.
                # # The time stamp stored for every entry is the insertion time
                # updatedMS = int( splitLine[3] )*1./1000
                # # convert to tuple format in UTC. Still to be checked which format the DB requires
                # updatedTuple = time.gmtime( updatedMS )
                # updated = time.asctime( updatedTuple )
                directories = filePath.split("/")
                fileName = directories[len(directories) - 1]
                basePath = ""
                for segment in directories[0:-1]:
                    basePath = basePath + segment + "/"
                if basePath not in self.dirDict:
                    self.dirDict[basePath] = {}
                    self.dirDict[basePath]["Files"] = 0
                    self.dirDict[basePath]["Size"] = 0
                    self.dirDict[basePath]["Updated"] = updated
                self.dirDict[basePath]["Files"] += 1
                self.dirDict[basePath]["Size"] += size
                processedLines += 1

            self.log.info("total lines: %d,  correctly processed: %d " % (totalLines, processedLines))
            # write to directory summary file
            fileP3 = outputFileMerged[st]["DirSummaryFileName"]
            fP3 = outputFileMerged[st]["pointerToDirSummaryFile"]
            self.log.info(f"Writing to file {fileP3}")
            for basePath in self.dirDict:
                summaryLine = " ".join(
                    [st, basePath, str(self.dirDict[basePath]["Files"]), str(self.dirDict[basePath]["Size"])]
                )
                self.log.debug(f"Writing summaryLine {summaryLine}")
                fP3.write(f"{summaryLine}\n")
            fP3.flush()
            fP3.close()

            self.log.debug("-------------------------summary of ReadInputFile-------------------------")
            for k in self.dirDict:
                self.log.debug(
                    "(lfn,st): %s-%s files=%d size=%d updated=%s"
                    % (k, st, self.dirDict[k]["Files"], self.dirDict[k]["Size"], self.dirDict[k]["Updated"])
                )

        return S_OK(retCode)

    def getLFNPath(self, site, pfnFilePath):
        """Given a PFN returns the LFN, stripping the suffix relative to the
        particular site.

        Important: usually the transformation is done simply removing the SApath of the site.
        So for ARCHIVE and FREEZER and FAILOVER data:
        the LFN will be: /lhcb/archive/<LFN> etc...
        even if LHCb register those replicas in the FC with the LFN: <LFN>, stripping the
        initial '/lhcb/archive'
        this is taken into account by the main method of the agent when it queries for replicas in the FC
        """

        outputFile = os.path.join(self.workDirectory, site + ".UnresolvedPFNs.txt")
        # this should be done with the replicaManager, but it does not work for archive files . tbu why
        # seName = site + '-RAW'
        sePath = self.siteConfig[site]["SAPath"]
        lfn = "None"
        try:
            lfn = pfnFilePath.split(sePath)[1]
        except BaseException:
            self.log.error(f"ERROR retrieving LFN from PFN = {pfnFilePath}, sePath = {sePath} ")
            if not os.path.exists(outputFile):
                of = open(outputFile, "w")
            else:
                of = open(outputFile, "a")
            of.write(f"{pfnFilePath}\n")
            of.close()
            return S_ERROR("Could not retrieve LFN")
        # additional check on the LFN format:
        if not lfn.startswith("/lhcb"):
            self.log.error(f"LFN should start with /lhcb: PFN={pfnFilePath} LFN={lfn}. Skip this file.")
            if not os.listdir(outputFile):
                of = open(outputFile, "w")
            else:
                of = open(outputFile, "a")
            os.write(f"{pfnFilePath}\n")
            os.close()
            return S_ERROR("Anomalous LFN does not start with '/lhcb' string")

        return S_OK(lfn)

    def urlretrieveTimeout(self, url, fileName, timeout=0):
        """Borrowed from dirac-install (and slightly modified to fit in this
        agent).

        Retrieve remote url to local file (fileName), with timeout wrapper
        """
        # NOTE: Not thread-safe, since all threads will catch same alarm.
        #       This is OK for dirac-install, since there are no threads.
        self.log.info(f'Retrieving remote file "{url}"')

        if timeout:
            signal.signal(signal.SIGALRM, alarmTimeoutHandler)
            # set timeout alarm
            signal.alarm(timeout)
        try:
            remoteFD = urlopen(url)
            expectedBytes = int(remoteFD.info()["Content-Length"])
            localFD = open(fileName, "wb")
            receivedBytes = 0
            data = remoteFD.read(16384)
            while data:
                receivedBytes += len(data)
                localFD.write(data)
                data = remoteFD.read(16384)
            localFD.close()
            remoteFD.close()
            if receivedBytes != expectedBytes:
                self.log.info(f"File should be {expectedBytes} bytes but received {receivedBytes}")
                return False
        except HTTPError as x:
            if x.code == 404:
                self.log.info(f"{url} does not exist")
                return False
        except Exception as x:
            if x == "TimeOut":
                self.log.info(f'Timeout after {str(timeout)} seconds on transfer request for "{url}"')
            if timeout:
                signal.alarm(0)
            raise x

        if timeout:
            signal.alarm(0)
        return True

    def downloadAndExtractTarball(self, originFileName, originURL, targetPath):
        """Borrowed from dirac-install ( slightly modified to fit in this agent).

        It download a tar archive and extract the content, using the method
        urlretrieveTimeout
        """
        tarName = f"{originFileName}"
        # destination file:
        tarPath = os.path.join(targetPath, tarName)
        try:
            if not self.urlretrieveTimeout(f"{originURL}/{tarName}", tarPath, 300):
                self.log.error(f"Cannot download {tarName}")
                return S_ERROR("Cannot download file")
        except Exception as e:
            self.log.error(f"Cannot download {tarName}: {str(e)}")
            return S_ERROR("Cannot download file")
        # check if the file has to be uncompressed
        self.log.info(f"The downloaded file is: {tarPath} ")
        if tarPath[-3:] not in ["bz2", "tgz", "tar"]:
            self.log.info("File ready to be read (no need to uncompress) ")
            return S_OK()
        # Extract
        cwd = os.getcwd()
        os.chdir(targetPath)
        try:
            tf = tarfile.open(tarPath, "r")
        except Exception as e:
            self.log.error(f"Cannot open file {tarPath}: {str(e)}")
            return S_ERROR("Cannot open file")
        for member in tf.getmembers():
            tf.extract(member)
        os.chdir(cwd)
        # Delete tar
        os.unlink(tarPath)
        return S_OK()

    def downloadFiles(self, originFileNames, originURL, targetPath):
        """Downloads a list of files from originURL locally to targetPath."""
        if not isinstance(originFileNames, list):
            self.log.error("first argument for downloadFiles method should be a list! ")
            return False
        for fileName in originFileNames:
            destinationPath = os.path.join(targetPath, fileName)
            try:
                if not self.urlretrieveTimeout(f"{originURL}/{fileName}", destinationPath, 300):
                    self.log.error(f"Cannot download {fileName}")
                    return False
            except Exception as e:
                self.log.error(f"Cannot download {fileName}: {str(e)}")
                return False
        return True

    def pathInLFC(self, dirName):
        """Get the path as registered in the FC.

        Different from the path that is used to build the pfn only for the
        special replicas (failover, archive, freezer)
        """
        lfcDirName = dirName
        for specialReplica in self.specialReplicas:
            prefix = "/lhcb/" + specialReplica
            if prefix in dirName:
                lfcDirName = dirName.split(prefix)[1]
                self.log.verbose(f"special replica! dirname = {dirName} -- lfcDirName = {lfcDirName}")
                return lfcDirName
        return lfcDirName

    def pathWithSuffix(self, dirName, replicaType):
        """Takes in input the path as registered in FC and returns the path with
        the initial suffix for the special replicas."""
        pathWithSuffix = dirName
        if replicaType in self.specialReplicas:
            pathWithSuffix = "/lhcb/" + replicaType + dirName
        return pathWithSuffix

    def getProblematicDirsSummary(self, site):
        """Produce a list of files that are not registered in the File Catalog and
        writes it down to a text file:

        1. queries the problematicDirs table to get all directories for a given site that have
           more data on SE than in LFCfor each replica type: (normal, archive, failover, freezer )
        2. scan the input files (from the sites storage dumps) to get all the files belonging
           to the problematic directories
        3. lookup in in FC file by file to check if they have a replica registered at the site
        4. the files that are found not to have a replica registered for the site, are written down to a file
        """
        self.log.info(f"*** Execute getProblematicDirsSummary method for site: {site} ")
        fileNameMissingReplicas = os.path.join(self.workDirectory, site + ".replicasMissingFromSite.txt")
        self.log.info(f"Opening file for replicas not registered: {fileNameMissingReplicas} ")
        fpMissingReplicas = open(fileNameMissingReplicas, "w")
        fileNameMissingFiles = os.path.join(self.workDirectory, site + ".filesMissingFromFC.txt")
        self.log.info(f"Opening file for files not registered: {fileNameMissingFiles} ")
        fpMissingFiles = open(fileNameMissingFiles, "w")
        problem = "NotRegisteredInFC"
        res = self.storageUsage.getProblematicDirsSummary(site, problem)
        if not res["OK"]:
            self.log.error(f"ERROR! {res}")
            return S_ERROR(res)
        val = res["Value"]
        problematicDirectories = {}  # store a list of directories for each replica type
        self.log.verbose("List of problematic directories: ")
        for row in val:
            # ('SARA', 'LHCb-Disk', 0L, 43L, '/lhcb/MC/2010/DST/00007332/0000/', 'NotRegisteredInFC','failover')
            site = row[0]
            lfcPath = row[4]
            problem = row[5]
            replicaType = row[6]
            pathWithSuffix = self.pathWithSuffix(lfcPath, replicaType)
            self.log.verbose(f"{lfcPath} {replicaType} - {pathWithSuffix}")
            if replicaType not in problematicDirectories:
                problematicDirectories[replicaType] = []
            if pathWithSuffix not in problematicDirectories[replicaType]:
                problematicDirectories[replicaType].append(pathWithSuffix)  # fix 13.03.2012
            else:
                self.log.error(
                    "the directory should be listed only once for a given site and type of replica! "
                    " site=%s, path= %s, type of replica =%s  " % (site, lfcPath, replicaType)
                )
                continue
        self.log.info("Found the following problematic directories:")
        for replicaType, problematicDir in problematicDirectories.items():
            self.log.info(f"replica type: {replicaType} , directories: {problematicDir} ")
        # retrieve the list of files belonging to problematic directories from the merged files:
        filesInProblematicDirs = {}
        # read the files from the Merged files
        pathToMergedFiles = self.siteConfig[site]["pathToInputFiles"]
        if pathToMergedFiles[:-1] != "/":
            pathToMergedFiles = pathToMergedFiles + "/"
        for mergedFile in os.listdir(pathToMergedFiles):
            if "Merge" not in mergedFile:
                continue
            fullFilePath = os.path.join(pathToMergedFiles, mergedFile)
            self.log.info(f"Scanning file: {fullFilePath} ... ")
            for line in open(fullFilePath).readlines():
                lfn = line.split()[0]  # this LFN includes the initial suffix (e.g./lhcb/archive/)
                directories = lfn.split("/")
                basePath = ""
                for segment in directories[0:-1]:
                    basePath = basePath + segment + "/"  # basepath is the directory including the initial suffix
                for replicaType in problematicDirectories:
                    if basePath in problematicDirectories[replicaType]:  # these paths do include initial suffix
                        if replicaType not in filesInProblematicDirs:
                            filesInProblematicDirs[replicaType] = []
                        if lfn not in filesInProblematicDirs[replicaType]:  # lfn including suffix
                            filesInProblematicDirs[replicaType].append(lfn)

        self.log.info("Files in problematic directories:")
        for replicaType in filesInProblematicDirs:
            self.log.info(f"replica type: {replicaType} files: {len(filesInProblematicDirs[replicaType])} ")
            for fil in filesInProblematicDirs[replicaType]:
                self.log.verbose(f"file in probl Dir {fil} {replicaType}")

        for replicaType in filesInProblematicDirs:
            res = self.checkReplicasInFC(
                replicaType, filesInProblematicDirs[replicaType], site, fileNameMissingReplicas, fileNameMissingFiles
            )
        fpMissingReplicas.close()
        fpMissingFiles.close()
        if self.pathToUploadResults[-1] != "/":
            self.pathToUploadResults = self.pathToUploadResults + "/"
        lfnToUploadResults = self.pathToUploadResults + fileNameMissingReplicas.split("/")[-1]
        self.log.info(f"Upload the file {fileNameMissingReplicas} to the grid with LFN: {lfnToUploadResults} ")
        guid = None
        res = self.diracApi.removeFile(lfnToUploadResults)
        if not res["OK"]:
            self.log.error(f"Failed to remove the file already existing {res['Message']} ")
        else:
            res = self.diracApi.addFile(
                lfnToUploadResults, fileNameMissingReplicas, "CERN-DEBUG", guid, printOutput=True
            )
            if not res["OK"]:
                self.log.error(f"Failed to upload to the grid the file {fileNameMissingReplicas} : {res['Message']} ")
        lfnToUploadResults = self.pathToUploadResults + fileNameMissingFiles.split("/")[-1]
        self.log.info(f"Upload the file {fileNameMissingFiles} to the grid with LFN: {lfnToUploadResults} ")
        res = self.diracApi.removeFile(lfnToUploadResults)
        if not res["OK"]:
            self.log.error(f"Failed to remove the file already existing {res['Message']} ")
        else:
            res = self.diracApi.addFile(lfnToUploadResults, fileNameMissingFiles, "CERN-DEBUG", guid, printOutput=True)
            if not res["OK"]:
                self.log.error(f"Failed to upload to the grid the file {fileNameMissingFiles} : {res['Message']} ")
        return S_OK()

    # ...............................................................................................................
    def checkReplicasInFC(self, replicaType, filesToBeChecked, site, fileNameMissingReplicas, fileNameMissingFiles):
        """Check the existance of the replicas for the given site and replica type
        in the FC."""
        self.log.info(f"*** Execute checkReplicasInFC for replicaType={replicaType}, site={site} ")
        filesMissingFromFC = []
        replicasMissingFromSite = []
        # totalSizeMissingFromFC = 0
        # totalSizeReplicasMissingFromSite = 0
        # for files in problematic directories look up in the FC:
        specialReplicasSEs = []
        for sr in self.specialReplicas:
            se = site + "-" + sr.upper()
            specialReplicasSEs.append(se)
        self.log.verbose(f"SEs for special replicas: {specialReplicasSEs} ")

        filesInProblematicDirs = []
        if replicaType in self.specialReplicas:
            for lfn in filesToBeChecked:
                filesInProblematicDirs.append(self.pathInLFC(lfn))
        else:
            filesInProblematicDirs = filesToBeChecked

        active = False  # then this should be moved to a config parameter
        start = time.time()
        if not filesInProblematicDirs:
            self.log.info(f"No file to be checked in the FC for site {site} ")
            return S_OK()
        if active:
            repsResult = self.dataManager.getActiveReplicas(filesInProblematicDirs)
        else:
            repsResult = self.dataManager.getReplicas(filesInProblematicDirs)
        timing = time.time() - start
        self.log.info(
            "%d replicas Lookup Time: %.2f s -> %.2f s/replica"
            % (len(filesInProblematicDirs), timing, float(timing) / len(filesInProblematicDirs))
        )
        if not repsResult["OK"]:
            return S_ERROR(repsResult["Message"])
        goodFiles = repsResult["Value"]["Successful"]
        badFiles = repsResult["Value"]["Failed"]
        for lfn in badFiles:
            if "No such file or directory" in badFiles[lfn]:
                self.log.info(f"missing from FC {lfn} {replicaType}")
                # check if the storage files have been removed in the meanwhile after the storage dump creation and the check
                # to be done
                storageFileStatus = self.storageFileExists(lfn, replicaType, site)
                if storageFileStatus == 1:
                    self.log.info(f"Storage file exists: Inconsistent file! {lfn} ")
                    filesMissingFromFC.append(lfn)
                elif storageFileStatus == 0:
                    self.log.info(f"Storage file does not exist (temporary file) {lfn} ")
                else:
                    self.log.warn(f"Failed request for storage file {lfn} ")
            else:
                self.log.info(f"Unknown message from Fc: {lfn} - {badFiles[lfn]} ")
        for lfn in goodFiles:
            # check if the replica exists at the given site:
            replicaAtSite = False
            if replicaType in self.specialReplicas:
                specialReplicaSE = site + "-" + replicaType.upper()
                for se in goodFiles[lfn]:
                    if se == specialReplicaSE:
                        self.log.verbose(f"matching se: {se} ")
                        replicaAtSite = True
                        break
            else:
                for se in goodFiles[lfn]:
                    if se in specialReplicasSEs:
                        self.log.verbose(f"Replica type is {replicaType}, skip this SE: {se} ")
                        continue
                    if site in se:
                        self.log.verbose(f"matching se: {se} ")
                        replicaAtSite = True
                        break
            if not replicaAtSite:
                # check if storage file currently exists or if it was a temporary file
                storageFileStatus = self.storageFileExists(lfn, replicaType, site)
                if storageFileStatus == 1:
                    self.log.info(f"Storage file exists: Inconsistent file! {lfn} ")
                    replicasMissingFromSite.append(lfn)
                elif storageFileStatus == 0:
                    self.log.info(f"Storage file does not exist (temporary file) {lfn} ")
                else:
                    self.log.warn(f"Failed request for storage file {lfn} ")

        assignedSE = "Unknown"
        if replicaType in self.specialReplicas:
            assignedSE = site + "-" + replicaType.upper()

        fpMissingReplicas = open(fileNameMissingReplicas, "a")
        self.log.info(f"Writing list of replicas missing from site to file {fileNameMissingReplicas} ")
        for lfn in replicasMissingFromSite:
            fpMissingReplicas.write(f"{lfn} {assignedSE}\n")
            self.log.info(f"{lfn} {assignedSE}\n")
        fpMissingReplicas.flush()

        self.log.info(f"Writing list of files missing from FC to file {fileNameMissingFiles} ")
        fpMissingFiles = open(fileNameMissingFiles, "a")
        for lfn in filesMissingFromFC:
            fpMissingFiles.write(f"{lfn} {assignedSE}\n")
            self.log.info(f"{lfn} {assignedSE}\n")
        fpMissingFiles.flush()

        fileName = os.path.join(self.workDirectory, site + "." + replicaType + ".consistencyChecksSummary.txt")
        self.log.info(f"Writing consistency check summary to file {fileName} ")
        date = time.asctime()
        line = "Site: " + site + "  Date: " + date
        fp = open(fileName, "w")
        fp.write(f"{line}\n")
        totSizeMissingFromFC = 0  # to be implemented!!!
        totSizeReplicasMissingFromSite = 0  # to be implemented!!!
        fp.write(
            "Total number of LFN at site not registered in the FC: %d , "
            "total size: %.2f GB \n" % (len(filesMissingFromFC), totSizeMissingFromFC / 1.0e9)
        )
        fp.write(
            "Total number of replicas  at site not registered in the FC: %d , "
            "total size: %.2f GB \n" % (len(replicasMissingFromSite), totSizeReplicasMissingFromSite / 1.0e9)
        )
        fp.close()

        return S_OK()

    def storageFileExists(self, lfn, replicaType, site):
        """Check if the replica exists on storage. This is to filter many temporary
        files (e.g. un-merged..) that are removed in the while between storage dump
        and consistency check. Return values:

        -1 : request failed
         0 : storage file does not exist
         1 : storage file exists
        """
        storageFileExist = -1
        # get the PFN
        seList = []
        if replicaType in self.specialReplicas:
            seName = site + "-" + replicaType.upper()
            seList.append(seName)
        else:
            specialReplicasSEs = []
            for sr in self.specialReplicas:
                se = site + "-" + sr.upper()
                specialReplicasSEs.append(se)
            # seName is not known
            # try with all the SEs available for the site, except the ones for special replicas
            allSEs = self.siteConfig[site]["SEs"]
            for se in allSEs:
                if se not in specialReplicasSEs:
                    seList.append(se)
        self.log.verbose(f"list of SEs : {seList}")

        for seName in seList:
            storageElement = StorageElement(seName)
            res = storageElement.getURL(lfn)
            self.log.verbose(f"checking existance for {lfn} - {seName}")
            res = storageElement.exists(lfn)
            self.log.verbose(f"result of getStorageFileExists: {res} ")
            if not res["OK"]:
                self.log.error(f"error executing StorageElement.exists {res['Message']} ")
                storageFileExist = -1
                continue
            if res["Value"]["Failed"]:
                self.log.error(f"error executing StorageElement.exists {res} ")
                storageFileExist = -1
                continue
            elif res["Value"]["Successful"]:
                if res["Value"]["Successful"][lfn]:
                    self.log.info(f"storage file exists: {res['Value']} ")
                    storageFileExist = 1
                    return storageFileExist
                else:
                    self.log.verbose(f"storage file NOT found: {res['Value']} ")
                    storageFileExist = 0

        return storageFileExist

    def castorPreParser(self, site, inputFilesDir):
        """Preliminary parsing for Castor nameserver dump Separates the files in 3
        space tokens relying on the namespace."""

        if inputFilesDir[-1:] != "/":
            inputFilesDir = inputFilesDir + "/"
        inputFile = os.listdir(inputFilesDir)
        if len(inputFile) != 1:
            self.log.error(f"For this parser, there should be one and only one file. Found these files: {inputFile} ")
            return S_ERROR()

        p1FilesDict = {
            "LHCb_USER": {"fileName": inputFilesDir + "LHCb_USER.txt"},
            "LHCb-Disk": {"fileName": inputFilesDir + "LHCb-Disk.txt"},
            "LHCb-Tape": {"fileName": inputFilesDir + "LHCb-Tape.txt"},
        }
        for st in p1FilesDict:
            fp = open(p1FilesDict[st]["fileName"], "w")
            p1FilesDict[st]["filePointer"] = fp

        fullPath = os.path.join(inputFilesDir, inputFile[0])
        for line in open(fullPath).readlines():
            splitLine = line.split()
            if len(splitLine) < 2:
                self.log.debug(f"Empty directory. Skip line! {line} ")
                continue
            size = splitLine[4]
            pfn = splitLine[10]
            # updated = should be reconstructed on the basis of the ascii date
            update = "na"
            res = self.getLFNPath(site, pfn)
            if not res["OK"]:
                return S_ERROR()
            lfn = res["Value"]
            probableSpaceToken = ""
            if lfn[:11] == "/lhcb/user/":
                probableSpaceToken = "LHCb_USER"
            elif (lfn[-3:] == "raw" or lfn[-4:] == "sdst" or lfn[:19] == "/lhcb/archive/lhcb/") and (
                lfn[:20] != "/lhcb/failover/lhcb/"
            ):
                probableSpaceToken = "LHCb-Tape"
            else:
                probableSpaceToken = "LHCb-Disk"
            newLine = pfn + " |  " + size + " |  " + update
            fp = p1FilesDict[probableSpaceToken]["filePointer"]
            fp.write(f"{newLine}\n")

        for st in p1FilesDict:
            fp = p1FilesDict[st]["filePointer"]
            fp.close()
        return S_OK()

    def checkCreationDate(self, directory):
        """Check the storage dump creation date.

        Returns 0 if the creation date is more recent than a given time
        interval (set as configuration parameter), otherwise returns -1
        """
        retCode = 0

        for fileName in os.listdir(directory):
            fullPath = os.path.join(directory, fileName)
            # # ( mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime ) = os.stat( fullPath )
            mtime = os.path.getmtime(fullPath)
            self.StorageDumpCreationTime = datetime.utcfromtimestamp(mtime)
            self.log.info(f"in checkCreationDate StorageDumpCreationTime {self.StorageDumpCreationTime} ")
            now = time.time()
            elapsedTime = now - mtime
            self.log.info(
                "creation time: %s, elapsed time: %d h "
                "(max delay allowed : %d h ) "
                % (time.ctime(mtime), elapsedTime / 3600, self.maximumDelaySinceSD / 3600)
            )

            if elapsedTime > self.maximumDelaySinceSD:
                self.log.warn(
                    "storage dump creation date is older "
                    "than maximum limit! %s s ( %d h ) - %s "
                    % (self.maximumDelaySinceSD, self.maximumDelaySinceSD / 3600, fullPath)
                )
                retCode = -1
                return S_OK(retCode)
        return S_OK(retCode)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" Analyse XMLSummary module and PoolCatalog in order to monitor the files access
    We send data to the accounting (Site -> SE : fail/success)
"""

from collections import defaultdict

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.AccountingSystem.Client.Types.DataOperation import DataOperation
from DIRAC.Resources.Catalog.PoolXMLCatalog import PoolXMLCatalog
from LHCbDIRAC.Core.Utilities.XMLSummaries import XMLSummary, XMLSummaryError
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase


class AnalyseFileAccess(ModuleBase):
    """Analyzing the access with xroot"""

    def __init__(self, bkClient=None, dm=None):
        """Module initialization."""

        self.log = gLogger.getSubLogger("AnalyseFileAccess")
        super().__init__(self.log, bkClientIn=bkClient, dm=dm)

        self.XMLSummary = ""
        self.XMLSummary_o = None
        self.poolXMLCatName = ""
        self.poolXMLCatName_o = None
        self.dsc = None

    def _resolveInputVariables(self):
        """By convention any workflow parameters are resolved here."""

        super()._resolveInputVariables()
        super()._resolveInputStep()

        self.XMLSummary_o = XMLSummary(self.XMLSummary, log=self.log)
        self.dsc = self.workflow_commons["AccountingReport"]
        self.poolXMLCatName_o = PoolXMLCatalog(xmlfile=self.poolXMLCatName)

    def execute(
        self,
        production_id=None,
        prod_job_id=None,
        wms_job_id=None,
        workflowStatus=None,
        stepStatus=None,
        wf_commons=None,
        step_commons=None,
        step_number=None,
        step_id=None,
    ):
        """Main execution method.

        Here we analyse what is written in the XML summary and the pool XML, and send accounting
        """

        try:
            super().execute(
                production_id,
                prod_job_id,
                wms_job_id,
                workflowStatus,
                stepStatus,
                wf_commons,
                step_commons,
                step_number,
                step_id,
            )

            try:
                self._resolveInputVariables()
            except XMLSummaryError as e:
                if e.message == "XML Summary Not Available":
                    self.log.warn("XML summary not created, skipping this module")
                    return S_OK()
                else:
                    raise e

            self.log.info(f"Analyzing root access from {self.XMLSummary} and {self.poolXMLCatName}")

            accessAttempts = self._checkFileAccess(self.poolXMLCatName_o, self.XMLSummary_o)

            if not self._enableModule():
                self.log.info("Not enabled")
                return S_OK()

            for remoteSE, success in accessAttempts:
                oDataOperation = self.__initialiseAccountingObject(remoteSE, success)
                self.dsc.addRegister(oDataOperation)

            return S_OK()

        except Exception as e:  # pylint:disable=broad-except
            self.log.exception("Failure in AnalyseFileAccess execute module", lException=e)
            self.setApplicationStatus(e)
            return S_ERROR(str(e))

        finally:
            super().finalize()

    @staticmethod
    def _checkFileAccess(xmlCatalog, xmlSummary):
        """Given an xmlCatalog and an xmlSummary, check which were the successful and failed attempts
        to open remote root files

        For each attempts, we return a tuple (srcSE, flag) with the flag being true if the read was successful.

        :param xmlCatalog: instance of :py:class:`~LHCbDIRAC.Resources.Catalog.PoolXMLCatalog.PoolXMLCatalog`
        :param xmlSummary: instance of :py:class:`~LHCbDIRAC.LHCbDIRAC.Core.Utilities.XMLSummaries.XMLSummary`

        :returns: list of tuples (srcSE, successful flag)
        """

        # This will contain the list of tuples with the accesses and their status
        accessAttempts = []

        # Used to cache the information PFN -> SE from the catalog
        pfn_se = {}
        # Used to cache the information LFN -> PFN from the catalog
        lfn_pfn = defaultdict(list)

        # Matching between a PFN and the LFN
        pfn_lfn = {}

        # {LFN: failed PFN}
        lfn_pfn_fail = defaultdict(list)

        # Build all the caches
        for guid in xmlCatalog.files:
            pFile = xmlCatalog.files[guid]
            lfn = pFile.lfns[0]  # there can be only one
            for pfn, _ftype, se in pFile.pfns:
                pfn_lfn[pfn] = lfn
                pfn_se[pfn] = se
                lfn_pfn[lfn].append(pfn)

        # Check the URLs that failed, and associate them with their LFNs
        for pfn, _status in xmlSummary.failedInputURL:
            # The name starts with 'PFN:'
            pfn = pfn[4:]
            # Find the matching LFN
            lfn = pfn_lfn[pfn]
            lfn_pfn_fail[lfn].append(pfn)

        # For each PFN in the lfn_pfn_fail dictionnary, we will count a failure.
        for lfn, failedPfns in lfn_pfn_fail.items():
            # We add the accounting for the failure
            for pfn in failedPfns:
                remoteSE = pfn_se[pfn]
                accessAttempts.append((remoteSE, False))

        # To find the successful access, we take the next available replicas if any.
        # If the LFN is not in the lfn_pfn_fail dict, that means the first attempt was successful
        # If there are no more replicas to be tried, then the LFN was never read.

        for lfn, replicaList in lfn_pfn.items():
            # get the index of the successful replicas
            succRepIndex = len(lfn_pfn_fail.get(lfn, []))
            try:
                succRep = replicaList[succRepIndex]
                remoteSE = pfn_se[succRep]
                accessAttempts.append((remoteSE, True))
            # This happens if all the replicas failed
            except IndexError:
                pass

        return accessAttempts

    def __initialiseAccountingObject(self, srcSE, successful):
        """create accouting record"""
        accountingDict = {}

        accountingDict["OperationType"] = "fileAccess"
        accountingDict["User"] = self._getCurrentOwner()
        accountingDict["Protocol"] = "xroot"
        accountingDict["RegistrationTime"] = 0.0
        accountingDict["RegistrationOK"] = 0
        accountingDict["RegistrationTotal"] = 0
        accountingDict["Source"] = srcSE
        accountingDict["TransferTotal"] = 1
        accountingDict["TransferOK"] = 1 if successful else 0
        accountingDict["TransferSize"] = 0
        accountingDict["TransferTime"] = 0.0
        accountingDict["FinalStatus"] = "Successful" if successful else "Failed"
        accountingDict["Destination"] = self.siteName
        oDataOperation = DataOperation()
        oDataOperation.setValuesFromDict(accountingDict)

        if "StartTime" in self.step_commons:
            oDataOperation.setStartTime(self.step_commons["StartTime"])
            oDataOperation.setEndTime()

        return oDataOperation



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Reads .log-files and outputs summary of counters as a .json-file
"""
import os
import json

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities import TimeUtilities
from DIRAC.Core.Utilities.ReturnValues import DReturnType


def readLogFile(
    logFile: str,
    jobID: str,
    prodID: str,
    wmsID: str,
    application: str,
    applicationVersion: str,
    name: str = "errors.json",
) -> DReturnType:
    """The script that runs everything.

    :param logFile: the name of the logfile
    :param jobID: the JobID
    :param prodID: the production ID
    :param wmsID: the wmsID
    :param name: the name of the output json file, standardised to 'errors.json'
    """
    logString = ""
    dictG4Errors = {
        "G4Exception": "",
        "G4 Exception": "",
        "ERROR ": "",
        "FATAL ": "",
        "PYTHIA WARNING ": "",
    }
    errorG4Dict = dict()
    errorDict = dict()
    if logFile.endswith(".log"):
        res = __getLogString(logFile, logString)
        if not res["OK"]:
            return res
        logString = res["Value"]
    else:
        gLogger.debug("The log is already in a readable string")
        logString = logFile

    reversedKeys = sorted(dictG4Errors, reverse=True)
    for errorString in reversedKeys:
        ctest = logString.count(errorString)
        test = logString.find(errorString)
        for i in range(ctest):
            start = test
            test = logString.find(errorString, start)
            alreadyFound = False
            for error in reversedKeys:
                if error == errorString:
                    break
                checke = logString[test : test + 100].find(error)
                if checke != -1:
                    alreadyFound = True
                    test += len(error)
                    break
            if alreadyFound:
                continue

            if test != -1:
                if errorString.find("G4") != -1:
                    check = logString[test : test + 250].find("***")
                    if check != -1:
                        errorBase = logString[test : test + 250].split("***")[0]
                        strippedErrString = errorBase.rstrip().replace("\n", "")
                        if not strippedErrString.startswith("G4Exception-END") and not strippedErrString.startswith(
                            "G4Exception-START"
                        ):
                            if strippedErrString in errorG4Dict:
                                errorG4Dict[strippedErrString] = errorG4Dict[strippedErrString] + 1
                            else:
                                errorG4Dict[strippedErrString] = 1
                        test += len(errorBase)
                else:
                    errorBase = logString[test : test + 250].split("\n")[0].rstrip()
                    if errorBase in errorDict:
                        errorDict[errorBase] = errorDict[errorBase] + 1
                    else:
                        errorDict[errorBase] = 1
                    test += len(errorBase)

    dictBase = {
        "JobID": jobID,
        "ProductionID": prodID,
        "wmsID": wmsID,
        "Application": application,
        "ApplicationVersion": applicationVersion,
        "timestamp": int(TimeUtilities.toEpochMilliSeconds()),
    }
    with open(name, "w") as output:
        output.write(json.dumps(dictBase | errorDict | errorG4Dict, indent=2))
    gLogger.notice("Finished creating the JSON file with Gauss Errors")
    return S_OK()


def __getLogString(logFile: str, logString: str) -> DReturnType:
    """Checks if the log file can be opened, and saves the text in logFile into logString.

    :param logFile: the name of the logFile
    :param logStr: the name of the variable that will save the contents of logFile
    """

    gLogger.notice("Attempting to open", logFile)
    if not os.path.exists(logFile):
        gLogger.error("File could not be found", f"({logFile})")
        return S_ERROR()
    if os.stat(logFile)[6] == 0:
        gLogger.error("File is empty", f"({logFile})")
        return S_ERROR()
    with open(logFile) as f:
        logString = f.read()
    gLogger.notice("Successfully read", logFile)
    return S_OK(logString)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""The LHCb AncestorFilesAgent queries the Bookkeeping catalogue for ancestor
files if the JDL parameter AncestorDepth is specified.

The ancestor files are subsequently added to the existing input data
requirement of the job.
"""
import time
from DIRAC import S_OK, S_ERROR
from DIRAC.WorkloadManagementSystem.Agent.OptimizerModule import OptimizerModule
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient


class AncestorFilesAgent(OptimizerModule):
    """Connects to BKK, ran through the optimizer."""

    def __init__(self, agentName, loadName, baseAgentName=False, properties={}):
        """c'tor."""
        OptimizerModule.__init__(self, agentName, loadName, baseAgentName, properties)
        self.bk = BookkeepingClient()

    #############################################################################
    def checkJob(self, job, classadJob):
        """The main agent execution method."""
        result = self.__checkAncestorDepth(job, classadJob)
        if not result["OK"]:
            return result

        return self.setNextOptimizer(job)

    #############################################################################
    def __checkAncestorDepth(self, job, classadJob):
        """This method checks the input data with ancestors.

        The original job JDL is always extracted to obtain the input data,
        therefore rescheduled jobs will not recursively search for ancestors
        of ancestors etc.
        """
        inputData = []
        if classadJob.lookupAttribute("InputData"):
            inputData = classadJob.getListFromExpression("InputData")

        if not classadJob.lookupAttribute("AncestorDepth"):
            self.log.warn(f"No AncestorDepth requirement found for job {job}")
            return S_ERROR("AncestorDepth Not Found")

        ancestorDepth = classadJob.getAttributeInt("AncestorDepth")

        if ancestorDepth == 0:
            return S_OK("Null AncestorDepth specified")

        self.log.info(
            f"Job {job} has {len(inputData)} input data files and specified ancestor depth of {ancestorDepth}"
        )
        result = self.__getInputDataWithAncestors(job, inputData, ancestorDepth)
        if not result["OK"]:
            return result

        newInputData = result["Value"]

        classadJob.insertAttributeVectorString("InputData", newInputData)
        newJDL = classadJob.asJDL()
        result = self.__setJobInputData(job, newJDL, newInputData)
        return result

    ############################################################################
    def __getInputDataWithAncestors(self, job, inputData, ancestorDepth):
        """Extend the list of LFNs with the LFNs for their ancestor files for the
        generation depth specified in the job JDL."""
        inputData = [i.replace("LFN:", "") for i in inputData]
        start = time.time()
        try:
            result = self.bk.getFileAncestors(inputData, ancestorDepth, replica=True)
        except Exception as x:
            self.log.warn(f"getFileAncestors failed with exception:\n{x}")
            return S_ERROR("getFileAncestors failed with exception")

        self.log.info(f"BK lookup time {time.time() - start:.2f} s")
        self.log.debug(result)
        if not result["OK"]:
            report = self.setJobParam(job, self.am_getModuleParam("optimizerName"), result["Message"])
            if not report["OK"]:
                self.log.warn(report["Message"])
            self.log.warn(result["Message"])
            return S_ERROR("No Ancestors Found For Input Data")

        ancestors = [anc["FileName"] for ancList in result["Value"]["Successful"].values() for anc in ancList]
        newInputData = ancestors + inputData
        param = f"{len(ancestors)} ancestor files retrieved from BK for depth {ancestorDepth}"

        report = self.setJobParam(job, self.am_getModuleParam("optimizerName"), param)
        if not report["OK"]:
            self.log.warn(report["Message"])

        return S_OK(newInputData)

    def __setJobInputData(self, job, jdl, inputData):
        """Sets the new job input data requirement including ancestor files."""
        inputData = [i.replace("LFN:", "") for i in inputData]

        result = self.jobDB.setInputData(job, inputData)
        if not result["OK"]:
            self.log.error(result["Message"])
            return S_ERROR("Setting New Input Data")

        result = self.jobDB.setJobJDL(job, jdl)
        if not result["OK"]:
            self.log.error(result["Message"])
            return S_ERROR("Setting New JDL")

        return S_OK("Job updated")



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""It is used to test the Bookkeeping utilities."""
import unittest

from LHCbDIRAC.BookkeepingSystem.DB.Utilities import enoughParams, _ONE, _TWO


class UtilitiesTestCase(unittest.TestCase):
    def test_enoughParams(self):
        result = enoughParams({})
        self.assertEqual(result, False)

        result = enoughParams({"ReplicaFlag": "Test", "Visible": "Test"})
        self.assertEqual(result, False)

        for i in _ONE:
            result = enoughParams({i: "Test"})
            self.assertEqual(result, False)

        for i in _TWO:
            for j in _TWO:
                result = enoughParams({i: "Test", j: "Test"})
                self.assertEqual(result, False)

        result = enoughParams({"ConfigName": "Test", "ConfigVersion": "Test", "Production": "Test"})
        self.assertEqual(result, True)

        result = enoughParams({"Production": "Test", "ReplicaFlag": "Test", "Visible": "Test"})
        self.assertEqual(result, True)


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(UtilitiesTestCase)
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: RAWIntegrityAgent

.. module: RAWIntegrityAgent

:synopsis: RAWIntegrityAgent determines whether RAW files in CASTOR were migrated correctly.
"""
# # imports
import datetime

# # from DIRAC
from DIRAC import S_OK

# # from Core
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData

# # from DMS
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Resources.Storage.StorageElement import StorageElement
from LHCbDIRAC.DataManagementSystem.DB.RAWIntegrityDB import RAWIntegrityDB
from DIRAC.Core.Utilities.List import breakListIntoChunks

AGENT_NAME = "DataManagement/RAWIntegrityAgent"


class RAWIntegrityAgent(AgentModule):
    """.. class:: RAWIntegirtyAgent.

    :param RAWIntegrityDB rawIntegrityDB: RAWIntegrityDB instance
    :param str gatewayUrl: URL to online RequestClient
    """

    def __init__(self, *args, **kwargs):
        """c'tor."""

        AgentModule.__init__(self, *args, **kwargs)

        self.rawIntegrityDB = None
        self.fileCatalog = None
        self.onlineRequestMgr = None

    def initialize(self):
        """agent initialisation."""

        self.rawIntegrityDB = RAWIntegrityDB()

        # The file catalog is used to register file once it has been transfered
        # But we want to register it in all the catalogs except the RAWIntegrityDB
        # otherwise it is register twice
        # We also remove the BK catalog because some files are not registered there
        # (detector calibration files for example). The real data are registered in
        # the bookeeping by the DataMover
        self.fileCatalog = FileCatalog(catalogs="FileCatalog")

        # This sets the Default Proxy to used as that defined under
        # /Operations/Shifter/DataManager
        # the shifterProxy option in the Configuration can be used to change this default.
        self.am_setOption("shifterProxy", "DataProcessing")

        return S_OK()

    def _checkMigrationStatus(self, filesMetadata, lfnsMetadata):
        """Check that the lfns in parameters are properly migrated, and compares
        the checksum between castor and the RAWIntegrityDB.

        :param filesMetadata: dict {lfn: se metadata}
        :param lfnsMetadata: dict {lfn: metadata in RAWIntegrityDB}

        :returns: True/False in successful, Failed in case of problem
        """
        ############################################################
        #
        # Determine the files that have been newly migrated and their success
        #

        successful = {}
        failed = {}
        for lfn, seMetadata in filesMetadata.items():
            isMigrated = seMetadata.get("Migrated", False)
            # If it is not migrated, go to the next one
            if not isMigrated:
                successful[lfn] = False
                continue
            else:
                self.log.info(f"{lfn} is copied.")
                castorChecksum = seMetadata["Checksum"]
                onlineChecksum = lfnsMetadata[lfn]["Checksum"]
                if castorChecksum.lower().lstrip("0") == onlineChecksum.lower().lstrip("0").lstrip("x"):
                    self.log.info(f"{lfn} migrated checksum match.")
                    successful[lfn] = True
                else:
                    errStr = "Migrated checksum mis-match.", "{} {} {}".format(
                        lfn,
                        castorChecksum.lstrip("0"),
                        onlineChecksum.lstrip("0").lstrip("x"),
                    )
                    self.log.error(errStr)
                    failed[lfn] = errStr

        return S_OK({"Successful": successful, "Failed": failed})

    def getNewlyCopiedFiles(self, activeFiles):
        """Get the list of files newly copied and those not yet copied.

        :param activeFiles: dict {lfn:RawIntegrityDB metadata} for files in Active status

        :return: tuple filesNewlyCopied, filesNotYetCopied (lfns lists)
        """
        # This is a list of all the lfns that we will have newly copied
        filesNewlyCopied = []
        # This are the lfns that are not yet copied
        filesNotYetCopied = []

        self.log.info("Obtaining physical file metadata.")
        # Group the lfns by SEs
        seLfns = {}
        for lfn, metadataDict in activeFiles.items():
            se = metadataDict["SE"]
            seLfns.setdefault(se, []).append(lfn)

        for se in sorted(seLfns):
            lfnList = seLfns[se]
            failedMetadata = {}
            successfulMetadata = {}
            res = StorageElement(se).getFileMetadata(lfnList)
            if not res["OK"]:
                errStr = "Failed to obtain physical file metadata."
                self.log.error(errStr, res["Message"])
                failedMetadata = {lfn: errStr for lfn in lfnList}
            else:
                successfulMetadata = res["Value"]["Successful"]
                failedMetadata = res["Value"]["Failed"]

            if failedMetadata:
                self.log.info(f"Failed to obtain physical file metadata for {len(failedMetadata)} files.")

            if successfulMetadata:
                self.log.info(f"Obtained physical file metadata for {len(successfulMetadata)} files.")

                ############################################################
                #
                # Determine the files that have been newly migrated and their success
                #
                res = self._checkMigrationStatus(successfulMetadata, activeFiles)
                if not res["OK"]:
                    self.log.error("Error when checking migration status", res)
                else:
                    succCompare = res["Value"]["Successful"]
                    failedCompare = res["Value"]["Failed"]
                    seFilesCopied = []
                    seFilesNotCopied = []
                    # The copied files are those in True in the successful dictionary
                    for lfn, isCopied in succCompare.items():
                        if isCopied:
                            seFilesCopied.append(lfn)
                        else:
                            seFilesNotCopied.append(lfn)

                    filesNewlyCopied.extend(seFilesCopied)
                    filesNotYetCopied.extend(seFilesNotCopied)

                    self.log.info(f"{len(seFilesCopied)} files newly copied at {se}.")
                    self.log.info(f"Found {len(failedCompare)} checksum mis-matches at {se}.")

        return filesNewlyCopied, filesNotYetCopied

    def registerCopiedFiles(self, filesNewlyCopied, copiedFiles, allUnmigratedFilesMeta):
        """Register successfuly copied files (newly, or in Copied status in the DB)
        in the DFC.

        :param filesNewlyCopied: [lfns] of files newly copied
        :param copiedFiles: {lfn:RIDb metadata} of files that were in Copied state.
        :param allUnmigratedFilesMeta: {lfn:RI Db metadata} for all lfns non migrated at
                                      the beginning of the loop.

        :return: {lfn:True} for successfuly registered lfns
        """
        if filesNewlyCopied or copiedFiles:
            self.log.info(
                "Attempting to register %s newly copied and %s previously copied files"
                % (len(filesNewlyCopied), len(copiedFiles))
            )
        else:
            self.log.info("No files to be registered")

        # Update copiedFiles to also contain the newly copied files
        copiedFiles.update({lfn: allUnmigratedFilesMeta[lfn] for lfn in filesNewlyCopied})

        successfulRegister = {}
        failedRegister = {}

        # Try to register them by batch
        for lfnChunk in breakListIntoChunks(copiedFiles, 100):
            # Add the metadata
            lfnDictChuck = {lfn: copiedFiles[lfn] for lfn in lfnChunk}
            res = self.fileCatalog.addFile(lfnDictChuck)

            if not res["OK"]:
                self.log.error("Completely failed to register some successfully copied file.", res["Message"])
                failedRegister.update({lfn: res["Message"] for lfn in lfnDictChuck})
            else:
                successfulRegister.update(res["Value"]["Successful"])
                failedRegister.update(res["Value"]["Failed"])

        for lfn, reason in failedRegister.items():
            self.log.error("Failed to register lfn. Setting to Copied", f"{lfn}: {reason}")
            res = self.rawIntegrityDB.setFileStatus(lfn, "Copied")
            if not res["OK"]:
                self.log.error("Error setting file status to Copied", f"{lfn}: {res['Message']}")

        for lfn in successfulRegister:
            self.log.info(f"Successfully registered {lfn} in the File Catalog.")

        return successfulRegister

    def removeRegisteredFiles(self, filesNewlyRegistered, registeredFiles, allUnmigratedFilesMeta):
        """Remove successfuly registered files (newly, or in Registered status in
        the DB) from the OnlineStorage.

        :param filesNewlyCopied: [lfns] of files newly copied
        :param copiedFiles: {lfn:RIDb metadata} of files that were in Copied state.
        :param allUnmigratedFilesMeta: {lfn:RI Db metadata} for all lfns non migrated at
                                      the beginning of the loop.

        :return: {lfn:True} for successfuly registered lfns
        """
        if filesNewlyRegistered or registeredFiles:
            self.log.info(
                "Attempting to remove %s newly registered and %s previously registered files"
                % (len(filesNewlyRegistered), len(registeredFiles))
            )
        else:
            self.log.info("No files to be removed")

        # Update registeredFiles to also contain the newly registered files
        registeredFiles.update({lfn: allUnmigratedFilesMeta[lfn] for lfn in filesNewlyRegistered})

        onlineSE = StorageElement("OnlineRunDB")

        # Try to them them all
        res = onlineSE.removeFile(registeredFiles)

        filesNewlyRemoved = {}
        failedRemove = {}
        if not res["OK"]:
            self.log.error("Completely failed to remove successfully registered files.", res["Message"])
            failedRemove = {lfn: res["Message"] for lfn in registeredFiles}
        else:
            filesNewlyRemoved = res["Value"]["Successful"]
            failedRemove = res["Value"]["Failed"]

        for lfn, reason in failedRemove.items():
            self.log.error("Failed to remove lfn. Setting to Registered", f"{lfn}: {reason}")
            res = self.rawIntegrityDB.setFileStatus(lfn, "Registered")
            if not res["OK"]:
                self.log.error("Error setting file status to Registered", f"{lfn}: {res['Message']}")

        for lfn in filesNewlyRemoved:
            self.log.info(f"Successfully removed {lfn} from the Online storage. Setting it to Done")
            res = self.rawIntegrityDB.setFileStatus(lfn, "Done")
            if not res["OK"]:
                self.log.error("Error setting file status to Done", f"{lfn}: {res['Message']}")

        return filesNewlyRemoved

    def execute(self):
        """execution in one cycle."""

        # Don't use the server certificate otherwise the DFC wont let us write
        gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "false")

        ############################################################
        #
        # Obtain the files which have not yet been migrated
        #
        self.log.info("Obtaining un-migrated files.")
        res = self.rawIntegrityDB.getUnmigratedFiles()
        if not res["OK"]:
            errStr = "Failed to obtain un-migrated files."
            self.log.error(errStr, res["Message"])
            return S_OK()

        # allUnmigratedFilesMeta contain all the files that are not yet
        # migrated (not copied, not registered, not removed), as well as their metadata
        allUnmigratedFilesMeta = res["Value"]
        self.log.info(f"Obtained {len(allUnmigratedFilesMeta)} un-migrated files.")
        if not allUnmigratedFilesMeta:
            return S_OK()

        # activeFiles contains the files that are not yet copied
        activeFiles = {}
        # copiedFiles contains files that are copied but not yet registered
        copiedFiles = {}
        # registeredFiles contains files that are copied, registered, but not removed from source
        registeredFiles = {}

        # Assign them
        for lfn, lfnMetadata in allUnmigratedFilesMeta.items():
            status = lfnMetadata.pop("Status")
            if status == "Active":
                activeFiles[lfn] = lfnMetadata
            elif status == "Copied":
                copiedFiles[lfn] = lfnMetadata
            elif status == "Registered":
                registeredFiles[lfn] = lfnMetadata

        totalSize = 0
        for lfn, fileDict in activeFiles.items():
            totalSize += int(fileDict["Size"])

        ############################################################
        #
        # Checking newly copied files
        #

        # Get the list of lfns properly copied and not copied
        filesNewlyCopied, filesNotYetCopied = self.getNewlyCopiedFiles(activeFiles)

        ####################################################
        #
        # Registering copied files
        #
        ####################################################

        filesNewlyRegistered = self.registerCopiedFiles(filesNewlyCopied, copiedFiles, allUnmigratedFilesMeta)

        ####################################################
        #
        # Performing the removal from the online storage
        #
        ####################################################
        filesNewlyRemoved = self.removeRegisteredFiles(filesNewlyRegistered, registeredFiles, allUnmigratedFilesMeta)

        # Doing some accounting

        migratedSize = sum(allUnmigratedFilesMeta[lfn]["Size"] for lfn in filesNewlyRemoved)

        # The number of files that we failed at migrating
        # is the number of files at the beginning, minus the one we processed completely
        # minus those that are not yet copied
        failedMigrated = len(allUnmigratedFilesMeta) - len(filesNewlyRemoved) - len(filesNotYetCopied)

        res = self.rawIntegrityDB.setLastMonitorTime()
        migratedSizeGB = migratedSize / (1024 * 1024 * 1024.0)
        self.log.info("TotMigratedSize", migratedSizeGB)
        self.log.info("NewlyMigrated", len(filesNewlyRemoved))
        self.log.info("TotMigrated", len(filesNewlyRemoved))
        self.log.info("FailedMigrated", failedMigrated)
        self.log.info("TotFailMigrated", failedMigrated)

        return S_OK()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Transformation Files state machine (LHCb specific)"""
from DIRAC.ResourceStatusSystem.PolicySystem.StateMachine import State
from LHCbDIRAC.ProductionManagementSystem.Utilities.StateMachine import LHCbStateMachine


class TransformationFilesStateMachine(LHCbStateMachine):
    """Implementation of the state machine for the TransformationFiles."""

    def __init__(self, state):
        """c'tor Defines the state machine transactions."""

        super().__init__(state)

        self.states = {
            "Unused-inherited": State(
                13, ["Unused", "Processed-inherited", "MaxReset-inherited"], defState="Unused-inherited"
            ),
            "Assigned-inherited": State(
                12, ["Unused", "Processed-inherited", "MaxReset-inherited"], defState="Assigned-inherited"
            ),
            "MaxReset-inherited": State(11, ["Unused"], defState="MaxReset-inherited"),
            "Processed-inherited": State(10),  # final state
            "Moved": State(9),  # final state
            "Removed": State(8),  # final state
            "MissingInFC": State(7),  # final state
            "NotProcessed": State(6, ["Unused"], defState="NotProcessed"),
            "ProbInFC": State(5),  # final state
            "MaxReset": State(4, ["Removed", "Moved", "Problematic"], defState="MaxReset"),
            "Problematic": State(3, ["Removed", "Unused"], defState="Problematic"),
            "Processed": State(2),  # final state
            "Assigned": State(1, ["Unused", "Processed", "MaxReset", "Problematic"], defState="Assigned"),
            "Unused": State(
                0,
                ["Assigned", "MissingInFC", "ProbInFC", "Problematic", "Removed", "NotProcessed", "Processed", "Moved"],
                defState="Unused",
            ),
        }



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Find files in an Analysis Production containing a specified run number.
"""
from __future__ import annotations

import json
from datetime import datetime, timezone
import re
from itertools import islice
from pathlib import Path
import sys

from rich.progress import track

from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise, convertToReturnValue
from DIRAC.ConfigurationSystem.Client.ConfigurationClient import ConfigurationClient
from LHCbDIRAC.ProductionManagementSystem.Client.AnalysisProductionsClient import AnalysisProductionsClient
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

from LHCbDIRAC.ProductionManagementSystem.scripts.dirac_ap_input_metadata import (
    _get_info_from_apd,
    _get_ancestors_info,
    _attach_ancestors,
    _batched,
)


def parseArgs():
    version = ""
    output = ""

    @convertToReturnValue
    def setVersion(s: str):
        nonlocal version
        version = s

    @convertToReturnValue
    def setOutput(s: str):
        nonlocal output
        output = s

    switches = [
        ("", "version=", "Version of the Analysis Production (optional)", setVersion),
        ("", "output=", "Path to a file to dump the output to", setOutput),
    ]
    Script.registerSwitches(switches)
    Script.registerArgument("run: run to find files for")
    Script.registerArgument("wg: Name of the WG the production belongs to e.g. `bnoc`")
    Script.registerArgument("analysis: Name of the analysis e.g. `bds2kstkstb`")
    Script.registerArgument("name: Name of the job in the production e.g. `13104007_bs2kst_892_kstb_892_2011_magup`")
    Script.parseCommandLine(ignoreErrors=False)
    run, wg, analysis, name = Script.getPositionalArgs(group=True)
    wg = wg.lower()
    analysis = analysis.lower()
    name = name.lower()
    run = int(run)

    if not ConfigurationClient().ping()["OK"]:
        gLogger.fatal("Failed to contact CS, do you have a valid proxy?")
        sys.exit(1)

    return (
        run,
        wg,
        analysis,
        name,
        version,
        output,
    )


@Script()
def main():
    run, wg, analysis, name, version, output = parseArgs()
    config = {
        "wg": wg,
        "analysis": analysis,
        "name": name,
    }
    if version != "":
        config["version"] = version
    save_output = False
    if output != "":
        output = Path(output)
        save_output = True

    tc = TransformationClient()
    bk = BookkeepingClient()
    apc = AnalysisProductionsClient()

    gLogger.info("Finding Analysis Productions files")
    apd_info = _get_info_from_apd(apc, config)

    gLogger.info("Finding descendents")
    apd_id, filetype = _get_prodid_and_filetype(bk, tc, apd_info)
    collision_files = _get_collision_files(bk, apd_info, filetype, run)

    decs_dict = returnValueOrRaise(
        bk.getFileDescendents(collision_files, depth=apd_info["depth"], checkreplica=True, production=apd_id)
    )

    if decs_dict["Failed"] != []:
        raise RuntimeError("Failed to find descendents")

    lfns_from_run = []
    for lfns in decs_dict["Successful"].values():
        lfns_from_run += lfns

    lfns_from_run = sorted(set(lfns_from_run))

    if lfns_from_run != []:
        gLogger.always(f"Found {len(lfns_from_run)} files in Aprod from run {run}:")
        for lfn in lfns_from_run:
            gLogger.always(lfn)
    else:
        gLogger.warn(f"Could not find any files in Aprod from run {run}")

    if save_output:
        gLogger.always(f"Saving LFNs to {str(output)}")
        with open(output, "w") as out_file:
            for lfn in lfns_from_run:
                out_file.write(f"{lfn}\n")


def _get_prodid_and_filetype(bk, tc, apd_info):
    prev_tid = apd_info["transformations"][0]["id"]
    next_tid = apd_info["transformations"][1]["id"]
    query = None
    prev_query = returnValueOrRaise(tc.getBookkeepingQuery(prev_tid))
    next_query = returnValueOrRaise(tc.getBookkeepingQuery(next_tid))
    if "ProductionID" in prev_query and prev_query["ProductionID"] == next_tid:
        query = next_query
    elif "ProductionID" in next_query and next_query["ProductionID"] == prev_tid:
        query = prev_query
    else:
        raise RuntimeError(
            f"Could not determine the merge step for\
                           {apd_info['transformations']}"
        )
    filetype = query["FileType"]
    return next_tid, filetype


def _get_collision_files(bk, apd_info, filetype, run):
    test_lfn = apd_info["apd_lfns"][0]
    ancestors = returnValueOrRaise(bk.getFileAncestors(test_lfn, depth=apd_info["depth"]))
    if ancestors["Failed"] != []:
        raise FileNotFoundError(f"Could not find ancestors for lfn {test_lfn}")
    job_info = returnValueOrRaise(bk.getJobInfo(ancestors["Successful"][test_lfn][0]["FileName"]))
    prod_id = job_info[0][18]

    collision_query = {"RunNumber": run, "Production": prod_id, "FileType": filetype.upper(), "Visible": "Yes"}
    collision_files = returnValueOrRaise(bk.getFiles(collision_query))
    return collision_files


if __name__ == "__main__":
    main()
